"""Common utils.

"""

__author__ = 'Md Jahidul Hamid <jahidulhamid@yahoo.com>'
__copyright__ = 'Copyright © Md Jahidul Hamid <https://github.com/neurobin/>'
__license__ = '[BSD](http://www.opensource.org/licenses/bsd-license.php)'
__version__ = '0.0.1'


import importlib.util
import sys, os, re
from importlib import import_module as system_import_module

def Open(path: str, mode: str, **kwargs):
    """Wrapper for open with utf-8 encoding

    Args:
        path (str): path to file
        mode (str): file open mode

    Returns:
        open: open context manager handle
    """
    return open(path, mode, encoding='utf-8', **kwargs)

def import_from_path(name: str, path: str):
    """Import a module from path

    Args:
        name (str): module name (should be dotted path)
        path (str): path (file path)

    Returns:
        object: module object
    """
    spec = importlib.util.spec_from_file_location(name, path)
    module = importlib.util.module_from_spec(spec) # type: ignore
    sys.modules[name] = module
    spec.loader.exec_module(module) # type: ignore
    return module

def _cached_system_import(module_name):
    # Check whether module is loaded and fully initialized.
    if not (
        (module := sys.modules.get(module_name))
        and (spec := getattr(module, "__spec__", None))
        and getattr(spec, "_initializing", False) is False
    ):
        module = system_import_module(module_name)
    return module

def import_module(path, base_path=None):
    """Import a module by string path.

    Args:
        path (str): path to module, can be a dotted path as well
        base_path (str, optional): base path (file path only) from where it can be imported by absolute import. Defaults to None.

    Returns:
        module: imported module
    """
    absolute_unix_path = None
    if path[0] == os.path.sep and not base_path:
            absolute_unix_path = path
    if os.path.sep in path:
        path = path.replace(base_path, '') if base_path else path
        path = re.sub(r'\.py$', '', path)
        path = path.replace(os.path.sep, '.').strip('.')
    if absolute_unix_path:
        return import_from_path(path, absolute_unix_path)
    return _cached_system_import(path)

def import_object(path, object_name=None, base_path=None):
    """Import an object by string path.

    Args:
        path (str): path to object, can be a dotted path as well
        object_name (str, optional): name of the object to import. Defaults to last part of dotted path.
        base_path (str, optional): base path from where it can be imported by absolute import. Defaults to None.

    Returns:
        object: imported object
    """
    if not object_name:
        try:
            path, object_name = path.rsplit(".", 1)
        except ValueError as err:
            raise ImportError("%s doesn't look like an object path, please pass object_name parameter" % path) from err

    try:
        return getattr(import_module(path, base_path=base_path), object_name)
    except AttributeError as err:
        raise ImportError(
            'Module "%s" does not define a "%s" attribute/class'
            % (path, object_name)
        ) from err




import typing
from collections import OrderedDict
import re




# from pydantic import BaseModel
# from pydantic.dataclasses import dataclass, Field

# @dataclass
# class Users():
#     id: int = Field('id serial int primary key')
#     name: str = Field('varchar(45)')

#     class Config:
#         orm_mode = True

s = """
        class BigUser(User):
            name = Field("int")
            age = Field("int");zaz = Field("int")
            zzz = Field("int")
            profession\
            = Field('varchar(255)')

            bnb = Field("int")


"""

_fields_ = {
    'age': 23,
    'bnb': 23,
    'name': 'fds',
    'profession': 'Teacher',
    'zaz': 34,
    'zzz': 234,
}

def get_match_index(k, v):
    m = re.search(f"(;|^)[\\s]*\\b{k}\\s*=", s, re.M)
    if m:
        return m.start()
    return 0

res = get_match_index('zaz', 34)
print(res)

# def get_ordered_fields(fields, cls_def):
#     new_fields = OrderedDict()
#     for k, v in



import logging
import unittest

import morm.meta as mt


LOGGER_NAME = 'morm-test-field-'
log = logging.getLogger(LOGGER_NAME)



class TestMethods(unittest.TestCase):
    def test_Meta(self):
        with self.assertRaises(NotImplementedError):
            mta = mt.Meta()



if __name__ == "__main__":
    unittest.main(verbosity=2)



"""Common fields.
"""

__author__ = 'Md Jahidul Hamid <jahidulhamid@yahoo.com>'
__copyright__ = 'Copyright © Md Jahidul Hamid <https://github.com/neurobin/>'
__license__ = '[BSD](http://www.opensource.org/licenses/bsd-license.php)'
__version__ = '0.0.1'


import re
from morm.fields.field import Field
from morm.model import ModelType, ModelBase


class ForeignKey(Field):
    """Foreign key field.

    Args:
        model (ModelType): model class.
        on_delete (str, optional): query to add with ON DELETE. Either SET DEFAULT, SET NULL, RESTRICT, CASCADE or NO ACTION.
        on_update (str, optional): query to add With ON UPDATE.
        kwargs : See morm.fields.Field for details.
    """
    def __init__(self, model: ModelType, on_delete='', on_update='', **kwargs):
        fk_pk = model.Meta.pk
        fk_type = model.Meta._field_defs_[fk_pk].sql_type
        fk_table = model.Meta.db_table
        alter_q = 'ALTER TABLE "{table}" DROP CONSTRAINT IF EXISTS "__FK_{table}_{column}__";ALTER TABLE "{table}" ADD CONSTRAINT "__FK_{table}_{column}__" FOREIGN KEY ("{column}") REFERENCES "%s"("%s")' % (fk_table, fk_pk)
        if on_delete:
            alter_q += f' ON DELETE {on_delete}'
        if on_update:
            alter_q += f' ON UPDATE {on_update}'
        alter_q += ' DEFERRABLE INITIALLY DEFERRED;'
        super(ForeignKey, self).__init__(fk_type, **kwargs)
        pt = self.sql_conf.conf['sql_alter']
        self.sql_conf.conf['sql_alter'] = (alter_q, *pt)


class EmailField(Field):
    """Email field.

    Args:
        max_length (int, optional): Defaults to 255.
    """

    def __init__(self, max_length=255, **kwargs):
        sql_type = f'varchar({max_length})'
        def validate_email(email):
            EMAIL_VALIDATION_REGEX = re.compile(r'[^@\s]+@[^@\s]+\.[^@\s]+')
            return bool(EMAIL_VALIDATION_REGEX.fullmatch(email))
        super(EmailField, self).__init__(sql_type, validator=validate_email, **kwargs)




import unittest

from morm.version import __version__

class TestMethods(unittest.TestCase):
    def test_version(self):
        self.assertTrue(isinstance(__version__, str))
        self.assertTrue(len(__version__.split('.')) >= 3)

if __name__ == "__main__":
    unittest.main(verbosity=2)



"""Field class.
"""

__author__ = 'Md Jahidul Hamid <jahidulhamid@yahoo.com>'
__copyright__ = 'Copyright © Md Jahidul Hamid <https://github.com/neurobin/>'
__license__ = '[BSD](http://www.opensource.org/licenses/bsd-license.php)'
__version__ = '0.0.1'

import copy, inspect
from typing import Any, Optional, Callable, Tuple, Dict, List, Union
from morm.types import Void
import morm.exceptions as ex


def always_valid(value: Any)-> bool:
    """Always return True regradless of the value.

    Args:
        value (Any): value to be validated.

    Returns:
        bool: True
    """
    return True

def nomodify(value: Any) -> Any:
    """Return the value without any modification.

    Args:
        value (Any): Any value.

    Returns:
        Any: The save value that is passed.
    """
    return value


class ColumnConfig():
    """Config container for each column/feild.

    This class is easily derivable from a json config.
    """
    def __init__(self, **kwargs):
        self.conf: Dict[str, Union[str, Tuple[str, ...]]] = kwargs

    def __eq__(self, other: 'ColumnConfig') -> bool: # type: ignore
        return self.conf['sql_alter'] == other.conf['sql_alter']\
                and self.conf['sql_type'] == other.conf['sql_type']\
                and self.conf['sql_ondrop'] == other.conf['sql_ondrop']

    def __repr__(self):
        reprs = []
        for k, v in self.conf.items():
            reprs.append(f'{k}={repr(v)}')
        body = ', '.join(reprs)
        return f'{self.__class__.__name__}({body})'

    def to_json(self) -> Dict[str, Union[str, Tuple[str, ...]]]:
        return self.conf

    def get_query_column_create_stub(self, table_name: str) -> Tuple[str, str]:
        """Get create query along with alter query for this field

        Can be used to make the create table query.

        Returns:
            Tuple[str, str]: create_query, alter_query
        """
        create_q = f'"{self.conf["column_name"]}" {self.conf["sql_type"]} {self.conf["sql_onadd"]}'
        alter_q, msg = self.get_query_column_alter((), table_name)
        return create_q, alter_q


    def get_query_column_add(self, table_name: str) -> Tuple[str, str]:
        """Get a sql query to add the column.

        Returns:
            Tuple[str, str]: sql query, message
        """
        queries = []
        msgs = []
        queries.append(f'ALTER TABLE "{table_name}" ADD COLUMN "{self.conf["column_name"]}" {self.conf["sql_type"]} {self.conf["sql_onadd"]};')
        msgs.append(f'\n* > ADD: {self.conf["column_name"]}: {self.conf["sql_type"]}')

        q, m = self.get_query_column_alter((), table_name)
        if q:
            queries.append(q)
            msgs.append(m)

        if self.conf['sql_engine'] == 'postgresql':
            return '\n'.join(queries), ''.join(msgs)
        else:
            raise ex.UnsupportedError(f"{self.conf['sql_engine']} not supported yet.")

    def get_query_column_drop(self, table_name: str) -> Tuple[str, str]:
        """Get a sql query to drop the column.

        Returns:
            Tuple[str, str]: sql query, message
        """
        query = f'ALTER TABLE "{table_name}" DROP COLUMN "{self.conf["column_name"]}" {self.conf["sql_ondrop"]};'
        msg = f'\n* > DROP: {self.conf["column_name"]} {self.conf["sql_ondrop"]}'
        if self.conf['sql_engine'] == 'postgresql':
            return query, msg
        else:
            raise ex.UnsupportedError(f"{self.conf['sql_engine']} not supported yet.")

    def get_query_column_rename(self, old_column_name: str, table_name: str) -> Tuple[str, str]:
        """Get a sql query to rename the column.

        Args:
            old_column_name (str)

        Returns:
            Tuple[str, str]: sql query, message
        """
        query = f'ALTER TABLE "{table_name}" RENAME COLUMN "{old_column_name}" TO "{self.conf["column_name"]}";'
        msg = f"\n* > RENAME: {old_column_name} --> {self.conf['column_name']}"
        if self.conf['sql_engine'] == 'postgresql':
            return query, msg
        else:
            raise ex.UnsupportedError(f"{self.conf['sql_engine']} not supported yet.")

    def get_query_column_modify(self, prev: 'ColumnConfig', table_name: str) -> Tuple[str, str]:
        """Get a sql query to modify data type of the column.

        Args:
            prev (ColumnConfig): previous column config

        Returns:
            Tuple[str, str]: sql query, message
        """
        # TODO: handle alter settings.
        queries = []
        msgs = []
        if self.conf['sql_type'] != prev.conf['sql_type']:
            q = f'ALTER TABLE "{table_name}" ALTER COLUMN "{self.conf["column_name"]}" SET DATA TYPE {self.conf["sql_type"]};'
            queries.append(q)
            msg = f"\n* > MODIFY: {prev.conf['column_name']}: {prev.conf['sql_type']} --> {self.conf['sql_type']}"
            msgs.append(msg)

        settings_query, msg = self.get_query_column_alter(prev.conf['sql_alter'], table_name) # type: ignore
        if settings_query:
            queries.append(settings_query)
            msgs.append(msg)


        if self.conf['sql_engine'] == 'postgresql':
            return '\n'.join(queries), ''.join(msgs)
        else:
            raise ex.UnsupportedError(f"{self.conf['sql_engine']} not supported yet.")

    def get_query_column_alter(self, prev_sql_alter: Tuple[str, ...], table_name: str) -> Tuple[str, str]:
        """Get a sql query to apply the sql_alter settings comparing with
        another config: prev.

        Args:
            prev_sql_alter (Tuple[str]): previous sql_alter

        Returns:
            Tuple[str, str]: sql query, message
        """
        if not isinstance(prev_sql_alter, (tuple, list,)):
            raise TypeError(f"sql_alter {prev_sql_alter} must be a tuple, {type(prev_sql_alter)} given")
        if not isinstance(self.conf['sql_alter'], (tuple, list,)):
            raise TypeError(f"sql_alter {self.conf['sql_alter']} must be a tuple, {type(self.conf['sql_alter'])} given")
        query = ''
        msgs = []
        query_stubs = []
        column_name = self.conf["column_name"]
        for qs in self.conf['sql_alter']:
            if qs not in prev_sql_alter:
                # new settings
                qs = qs.format(table=table_name, column=column_name)
                query_stubs.append(qs)
                msgs.append(f"\n* + {qs}")

        if query_stubs:
            query = '; '.join(query_stubs) + ';'

        if self.conf['sql_engine'] == 'postgresql':
            return query, ''.join(msgs)
        else:
            raise ex.UnsupportedError(f"{self.conf['sql_engine']} not supported yet.")


class Field(object):
    """Initialize the Field object with data type (sql).

    Example sql_type: `'varchar(255)'`, `'integer'`, etc..

    Example sql_onadd: `'PRIMARY KEY'`, `'NOT NULL'`, `'UNIQUE'` etc..
    `sql_onadd` is only for adding the column, thus a change to this
    parameter will not trigger any change for the field.

    Example sql_ondrop: `'CASCADE'` and `'RESTRICT'`

    Example sql_alter queries

    ```sql
    ALTER TABLE "{table}" ALTER COLUMN "{column}" SET DEFAULT expression
    ALTER TABLE "{table}" ALTER COLUMN "{column}" DROP DEFAULT
    ALTER TABLE "{table}" ALTER COLUMN "{column}" SET NOT NULL
    ALTER TABLE "{table}" ALTER COLUMN "{column}" SET STATISTICS integer
    ALTER TABLE "{table}" ALTER COLUMN "{column}" SET ( attribute_option = value [, ... ] )
    ALTER TABLE "{table}" ALTER COLUMN "{column}" RESET ( attribute_option [, ... ] )
    ALTER TABLE "{table}" ALTER COLUMN "{column}" SET STORAGE { PLAIN | EXTERNAL | EXTENDED | MAIN }
    ```

    {table} and {column} will be replaced with table name,
    and column name respectively.


    Args:
        sql_type (str): Data type in SQL.
        sql_onadd (str): sql to add in ADD clause after 'ADD COLUMN column_name data_type'
        sql_ondrop (str): Either 'RESTRICT' or 'CASCADE'.
        sql_alter (Tuple[str]): Alter column queries; Example: ('ALTER TABLE "{table}" ALTER COLUMN "{column}" DROP DEFAULT',).
        sql_engine (str): db engine, postgresql, mysql etc.. Defaults to 'postgresql'
        default (Any, optional): Pythonic default value (can be a callable). Defaults to Void. (Do not use mutable values, use function instead)
        value (Any, optional): Set a value that will prevail unless changed manually. Can be a function. Useful to make updated_at like fields.
        choices (Tuple[Tuple[str, Any], ...], optional): choices tuple: (('Name of choice', 'value'), ...)
        help_text (str, optional):  help text to describe this field.
        validator (callable, optional): A callable that accepts exactly one argument. Validates the value in `clean` method. Defaults to always_valid.
        modifier (callable, optional): A callable that accepts exactly one argument. Modifies the value if validation fails when the `clean` method is called.. Defaults to nomodify.
        fallback (bool, optional): Whether invalid value should fallback to default value suppressing exception. (May hide bugs in your program)
    """
    def __init__(self, sql_type: str,
                sql_onadd='',
                sql_ondrop='',
                sql_alter: Tuple[str, ...] = (),
                sql_engine='postgresql',
                default: Any=Void,
                value: Any=Void,
                unique=False,
                choices: Tuple[Tuple[str, Any], ...] = (),
                help_text: str = '',
                validator: Callable=always_valid,
                modifier: Callable=nomodify,
                fallback=False,): # if you add new param here, update __repr__ method
        # Rules for using a variable name here as local variables go into the self._json_:
        # 1. Must precede with underscore if not in the parameter list
        # 2. Make sure to exclude unnecessary variables in the self._json_ like 'self' etc..
        _init_args = list(locals().keys())[1:] # this must be the first line here in __init__

        self.sql_type = sql_type
        _unique_constraint = '__UNQ_{table}_{column}__'
        if unique:
            _sql_unique = 'ALTER TABLE "{table}" ADD CONSTRAINT "%s" UNIQUE ("{column}");' % (_unique_constraint,)
        else:
            _sql_unique = 'ALTER TABLE "{table}" DROP CONSTRAINT IF EXISTS "%s";' % (_unique_constraint,)
        sql_alter = (_sql_unique, *sql_alter)
        self.sql_conf = ColumnConfig(sql_type=sql_type, sql_onadd=sql_onadd, sql_ondrop=sql_ondrop, sql_alter=sql_alter, sql_engine=sql_engine)
        self.validator = validator
        self.modifier = modifier
        self._name = ''
        self.fallback = fallback
        self._is_perpetual_default = False
        if value is not Void:
            self.default = value
            self._is_perpetual_default = True
            if default is not Void:
                raise ValueError(f"Parameter 'value' and 'default' both can not be set at the same time.")
        else:
            self.default = default
        self.choices = choices
        self.help_text = help_text

        # make a json
        self._json_ = {
            '_name': self.name,
            '_init_args': _init_args,
        }
        args = locals()
        for k,v in args.items():
            if k in ['self', '_init_args']: continue
            if callable(v):
                self._json_[k] = inspect.getsource(v)
            else:
                self._json_[k] = 'Void' if v == Void else v


    def __eq__(self, other: 'Field') -> bool: # type: ignore
        return other and self.sql_conf == other.sql_conf

    def __repr__(self):
        reprs = []
        for k in self._json_['_init_args']:
            if k == 'sql_type': reprs.append(repr(self.sql_type))
            elif k in ['sql_onadd','sql_ondrop','sql_alter','sql_engine',]:
                reprs.append(f'{k}={repr(self.sql_conf.conf[k])}')
            else:
                try:
                    if callable(self.__dict__[k]):
                        reprs.append(f'{k}={inspect.getsource(self.__dict__[k])}')
                    else:
                        reprs.append(f'{k}={repr(self.__dict__[k])}')
                except KeyError: pass
        body = ', '.join(reprs)
        return f'{self.__class__.__name__}({body})'

    def to_json(self):
        res = self._json_.copy()
        res['_name'] = self.name
        res['_repr'] = self.__repr__()
        del res['_init_args']
        return res

    def __invert__(self):
        return self.name

    @property
    def name(self) -> str:
        """Get the name of the field"""
        return self._name

    @name.setter
    def name(self, v: str):
        """Set the name of the field"""
        self._name = v
        self.sql_conf.conf['column_name'] = v

    def clean(self, value: Any, fallback: bool=False):
        """Clean the value by calling validator -> modifier -> validator

        If fallback is set to True, value that does not pass the
        validator will not raise any exception and will return the
        default value.


        Args:
            value (Any): Any value
            fallback (bool, optional): Fallback to default. Defaults to False.

        Raises:
            ValueError: If validation fails

        Returns:
            Any: value
        """
        if not self.validator(value):
            value = self.modifier(value)
        else:
            return value
        # the value may have been changed, try to validate again
        if not self.validator(value):
            if fallback:
                return self.get_default()
            raise ValueError("Value (%s) (type: %s) did not pass validation check for '%s'" % (str(value), type(value), self.name,))
        return value

    def get_default(self):
        """Get the default value.

        The default given is treated as a callable first, if it fails
        then it is returned as a value.

        Returns:
            Any: default value
        """
        try:
            return self.default()
        except TypeError:
            return self.default


class FieldValue():
    """Field Value Container

    `FieldValue().value` is a property that sets and gets the value.

    `FieldValue().value_change_count` gives the count of how many times the value
    has been set or changed.

    Args:
        field (Field): Field object
    """

    def __init__(self, field: Field):
        self._field = field
        self._value = Void
        self.value_change_count = 0
        self._ignore_first_change_count_ = False

    def __eq__(self, other):
        return self._field == other._field and self._value == other._value

    @property
    def value_change_count(self) -> int:
        """Return how many times the value has been set or changed.

        Returns:
            Any: value
        """
        return self.__value_change_count

    @value_change_count.setter
    def value_change_count(self, v: Any):
        """Set value change counter

        Args:
            v (Any): value
        """
        if not self._field._is_perpetual_default:
            self.__value_change_count = v
        else:
            self.__value_change_count = 1 # always change

    @property
    def value(self) -> Any:
        """Get the value if set or default otherwise.

        Returns:
            Any: value or default
        """
        if self._value is Void or (self._field._is_perpetual_default and self.value_change_count == 0):
            return self._field.get_default()
        return self._value

    @value.setter
    def value(self, v: Any):
        """Set the value by doing cleanup.

        Falls back to default if value is invalid.

        Args:
            v (Any): value
        """
        self.set_value(v, fallback=self._field.fallback)

    def set_value(self, v: Any, fallback=False):
        """Value setter.

        Args:
            v (Any): value
            fallback (bool, optional): Whether to fallback to default value if v is invalid. Defaults to False.
        """
        self._value = self._field.clean(v, fallback=fallback)
        if not self._ignore_first_change_count_: # if from db, then value change count won't change
            self.value_change_count += 1 # This must be the last line
        self._ignore_first_change_count_ = False
        # when the clean method raises exception, it will not be counted as
        # a successful value assignment.

    def delete_value(self):
        """Return the value to its initial state.
        """
        self._value = Void
        # must not change/decrease value_change_count
        # when this value is set again, the counter should increase
        # according to its previous value.
        # delete itself is a change but it does not change the value
        # to a valid value thus the counter should not increase either.



import logging
import unittest
import os, sys
import tempfile
import shutil
from morm import datetime


LOGGER_NAME = 'morm-test-admin-'
log = logging.getLogger(LOGGER_NAME)



class TestMethods(unittest.TestCase):
    def test_main(self):
        print(datetime.timestamp())
        print(datetime.timestampu())




if __name__ == "__main__":
    unittest.main(verbosity=2)




import unittest
from copy import copy, deepcopy
import pickle

from morm.types import Void, VoidType


class TestMethods(unittest.TestCase):
    def test_Void(self):
        self.assertTrue(Void is VoidType())
        self.assertTrue(len(Void) == 0)
        self.assertTrue(bool(Void) == False)
        self.assertTrue(Void is copy(Void))
        self.assertTrue(Void is deepcopy(Void))
        self.assertTrue(Void is VoidType())
        with self.assertRaises(NotImplementedError):
            Void.fdsfds = 9
        with self.assertRaises(NotImplementedError):
            Void['fds'] = 9

        s = pickle.dumps(Void)
        v = pickle.loads(s)
        self.assertTrue(v is Void)


if __name__ == "__main__":
    unittest.main(verbosity=2)



import logging
import unittest

from morm.q import Q


LOGGER_NAME = 'morm-test-q-'
log = logging.getLogger(LOGGER_NAME)



class TestMethods(unittest.TestCase):
    def test_Q(self):
        self.assertEqual(Q('string'), '"string"')



if __name__ == "__main__":
    unittest.main(verbosity=2)



"""Exceptions.
"""

__author__ = 'Md Jahidul Hamid <jahidulhamid@yahoo.com>'
__copyright__ = 'Copyright © Md Jahidul Hamid <https://github.com/neurobin/>'
__license__ = '[BSD](http://www.opensource.org/licenses/bsd-license.php)'
__version__ = '0.0.1'


class ItemDoesNotExistError(Exception): pass
class TransactionError(Exception): pass
class MigrationError(Exception): pass
class MigrationModelNotAllowedError(Exception): pass
class UnsupportedError(Exception): pass



"""Migrations utilities.
"""

__author__ = 'Md Jahidul Hamid <jahidulhamid@yahoo.com>'
__copyright__ = 'Copyright © Md Jahidul Hamid <https://github.com/neurobin/>'
__license__ = '[BSD](http://www.opensource.org/licenses/bsd-license.php)'
__version__ = '0.0.1'


import asyncio
import asyncpg # type: ignore
from typing import Dict, List, Tuple, Any, Union, Iterator, Callable, Optional
import re
import os, sys, shutil
import glob
import datetime
import json
from pathlib import Path
from morm.db import DB, ModelQuery, Transaction, Pool
from morm.model import ModelBase, ModelType
import morm.exceptions as exc
from morm.fields.field import ColumnConfig
from morm.utils import Open, import_from_path

HOME = str(Path.home())
MIGRATION_CURSOR_DIR = os.path.join(HOME, '.local', 'share', 'morm')
os.makedirs(MIGRATION_CURSOR_DIR, exist_ok=True)


def _get_changed_fields(curs: Dict[str, ColumnConfig],
                        pres: Dict[str, ColumnConfig])\
                        -> Dict[str, Dict[str, Union[None, str, ColumnConfig]]]:
    ops: Dict[str, Dict[str, Union[None, str, ColumnConfig]]] = {}
    cl = len(curs)
    pl = len(pres)
    bl = cl if cl >= pl else pl

    curs_k = iter(curs)
    pres_k = iter(pres)

    c = 0
    key_before = None
    while c < bl:
        c += 1 # loop control
        try:
            ck: Optional[str] = next(curs_k)
        except StopIteration:
            ck = None
        try:
            pk: Optional[str] = next(pres_k)
        except StopIteration:
            pk = None
        if ck:
            if ck in pres:
                # this field existed before
                if curs[ck] != pres[ck]:
                    # This field has been modified.
                    # Things can happen:
                    # 1. The field definition has been modified
                    # Can not handle the following:
                    # 2. The field was deleted and added a new (ignore)
                    # 3. The field was deleted and another one was renamed to this one.
                    # 4. The field was renamed and another one was renamed to this one.
                    ops[ck] = {
                        'op': 'mod',
                        'cur_key': ck,
                        'cur_def': curs[ck],
                        'pre_key': ck,
                        'pre_def': pres[ck],
                        'key_before': key_before,
                    }
            else:
                # This field did not exist before, this is
                # undoubtedly a new field. Either add or rename
                ops[ck] = {
                    'op': 'add',
                    'cur_key': ck,
                    'cur_def': curs[ck],
                    'key_before': key_before,
                }
        if pk:
            if pk in curs:
                # It will be handled by curs handler above.
                pass
            else:
                # pk is either deleted or renamed.
                if ck and ck not in pres:
                    # ck is new and at the position of pk
                    # take as renamed
                    ops[ck] = { # rename overlaps add in above.
                        'op': 'rename',
                        'cur_key': ck,
                        'cur_def': curs[ck],
                        'pre_key': pk,
                        'pre_def': pres[pk],
                        'key_before': key_before,
                    }
                    # if the values are not equal,
                    # then it can happen that pk was deleted and
                    # ck was added at its position, but it can
                    # also happen that pk was renamed and
                    # modified at the same time.
                    # Deleting will be loss of data, thus
                    # throwing error in renaming when the
                    # column definition does not comply
                    # seems much more sensible. One can
                    # do the changes one step at a time
                    # to avoid that error.
                # elif ck and ck in pres:
                #     # ck is not new and at the position of pk.
                #     # where pk no longer exists.
                #     pass
                else:
                    # pk is deleted.
                    ops[pk] = {
                        'op': 'delete',
                        'pre_key': pk,
                        'pre_def': pres[pk],
                        'key_before': key_before,
                    }

        key_before = pk
    return ops

MIGRATION_RUNNER_TEMPLATE = '''
import morm

class MigrationRunner(morm.migration.MigrationRunner):
    """Run migration with pre and after steps.
    """
    migration_query = """{migration_query}"""

    # async def run_before(self):
    #     """Run before migration

    #     self.tdb is the db handle (transaction)
    #     self.model is the model class
    #     """
    #     dbm = self.tdb(self.model)
    #     # # Example
    #     # dbm.q('SOME QUERY TO SET "column_1"=$1', 'some_value')
    #     # await dbm.execute()
    #     # # etc..

    # async def run_after(self):
    #     """Run after migration.

    #     self.tdb is the db handle (transaction)
    #     self.model is the model class
    #     """
    #     dbm = self.tdb(self.model)
    #     # # Example
    #     # dbm.q('SOME QUERY TO SET "column_1"=$1', 'some_value')
    #     # await dbm.execute()
    #     # # etc..
'''

class MigrationRunner():
    """Run migration with pre and after steps.
    """
    tdb: DB
    model: ModelType
    migration_query: str

    def __init__(self, tdb: DB, model: ModelType):
        self.tdb = tdb
        self.model = model

    async def run_before(self):
        """Run before migration

        self.tdb is the db handle (transaction)
        self.model is the model class
        """
        # dbm = self.tdb(self.model)
        # # Example
        # dbm.q('SOME QUERY TO SET "column_1"=$1', 'some_value')
        # await dbm.execute()
        # # etc..
        pass

    async def run_after(self):
        """Run after migration.

        self.tdb is the db handle (transaction)
        self.model is the model class
        """
        # dbm = self.tdb(self.model)
        # # Example
        # dbm.q('SOME QUERY TO SET "column_1"=$1', 'some_value')
        # await dbm.execute()
        # # etc..
        pass

    async def _run_migration_query(self):
        try:
            dbm = self.tdb(self.model)
            if self.migration_query:
                await dbm.q(self.migration_query).execute()
        except asyncpg.exceptions.PostgresSyntaxError:
            sys.stderr.write(f'************ Migration Query ************\n{dbm.getq()[0]}\n\n')

    async def run(self):
        """Runs run_before, _run_migration_query and run_after sequentially.
        """
        await self.run_before()
        await self._run_migration_query()
        await self.run_after()






class Migration():
    """Initiate a migration configuration.

    Args:
        model (ModelType): model class
        migration_base_path (str): where to save migration files.
        index_length (int, optional): index length of migration files. Defaults to 8.

    Raises:
        exc.MigrationModelNotAllowedError: For invalid model
    """
    def __init__(self, model: ModelType, migration_base_path: str, index_length=8):
        if model._is_abstract_():
            raise exc.MigrationModelNotAllowedError(f'Abstract model ({model.__name__}) can not be in database')
        if model._is_proxy_():
            raise exc.MigrationModelNotAllowedError(f"Proxy model ({model.__name__}) can not be passed for migration. Do migration with the non-proxy version.")
        self.model = model
        self.migration_base_path = migration_base_path
        self.index_length = index_length
        self.db_table = model._get_db_table_()
        self.migration_dir = os.path.join(migration_base_path, self.db_table)
        self.migration_queue_dir = os.path.join(self.migration_dir, '.queue')
        self.migration_file_pattern = os.path.join(self.migration_dir, f'{self.db_table}_*.json')
        os.makedirs(self.migration_dir, exist_ok=True)
        os.makedirs(self.migration_queue_dir, exist_ok=True)
        self.flatten_migration_dir = os.path.realpath(self.migration_dir).replace(os.path.sep, '_')
        self.migration_cursor_path = os.path.join(MIGRATION_CURSOR_DIR, f'{self.flatten_migration_dir}.cursor')

        self.previous_file_path = self._get_previous_migration_file_path()
        self.previous_file_name = os.path.basename(self.previous_file_path)
        self.current_file_name_without_extention = self._get_current_migration_file_name_without_extension(self.previous_file_name)
        self.current_file_name = f"{self.current_file_name_without_extention}.json"
        self.current_file_path = os.path.join(self.migration_dir, self.current_file_name)
        # self.current_sql_file_name = f'{self.current_file_name}.sql'
        # self.current_sql_file_path = os.path.join(self.migration_queue_dir, self.current_sql_file_name)
        self.current_mgrpy_file_name = f'{self.current_file_name_without_extention}.py'
        self.current_mgrpy_file_path = os.path.join(self.migration_queue_dir, self.current_mgrpy_file_name)
        self.mgrpy_file_pattern = os.path.join(self.migration_queue_dir, '*.py')

        self.default_json = {
            'db_table': self.db_table,
            'fields': {},
        }
        self._fields = self._get_fields()

        pjson = self._get_json_from_file(self.previous_file_path)
        fields = pjson['fields']
        self._pfields: Dict[str, ColumnConfig] = {}
        for k, v in fields.items():
            self._pfields[k] = ColumnConfig(**v)

        self.cjson_fields = {}
        for k, v in self.cfields.items():
            self.cjson_fields[k] = v.to_json()

        self.current_json = self.default_json
        self.current_json['fields'] = self.cjson_fields

    @property
    def cfields(self) -> Dict[str, ColumnConfig]:
        """ColumnConfig for all fields in a dict.

        Returns:
            Dict[str, ColumnConfig]
        """
        return self._fields

    @property
    def pfields(self) -> Dict[str, ColumnConfig]:
        """ColumnConfig for all previous fields in a dict.

        Returns:
            Dict[str, ColumnConfig]
        """
        return self._pfields

    def _get_fields(self) -> Dict[str, ColumnConfig]:
        fields = self.model._get_all_fields_()
        fieldscc: Dict[str, ColumnConfig] = {}
        for k, v in fields.items():
            fieldscc[k] = v.sql_conf
        return fieldscc

    @staticmethod
    def _get_create_table_query(db_table: str, fields: Dict[str, ColumnConfig]) -> str:
        """Get the complete create table query

        Args:
            fields (Dict[str, ColumnConfig]): fields that will be used to create the create quey

        Returns:
            str: query string
        """
        cq0 = f'CREATE TABLE "{db_table}" (\n'
        cq = []
        aq = []
        for k, v in fields.items():
            q, alq = v.get_query_column_create_stub(db_table)
            cq.append(f'    {q}')
            aq.append(alq)
        cqm = ',\n'.join(cq)
        cqe = '\n);'
        create_q = f'{cq0}{cqm}{cqe}'
        alter_q = '\n'.join(aq)
        total_q = f'{create_q}{alter_q}'
        return total_q

    def get_create_table_query(self) -> str:
        """Get a create table query for current fields

        Returns:
            str: query string
        """
        return self._get_create_table_query(self.db_table, self.cfields)

    def _get_json_from_file(self, path: str) -> Dict[str, Any]:
        """Get json from file or return a default json if file does not exist.

        Args:
            path (str): file path

        Returns:
            Dict[str, Any]: json dict
        """
        if not path:
            return self.default_json
        try:
            with Open(path, 'r') as f:
                return json.load(f)
        except FileNotFoundError:
            return self.default_json

    def _take_yn(self):
        yn = input("Is this correct? [Y/n]: ")
        if yn != 'Y' and yn != 'y': sys.exit()
        return True

    def _move_to_trash(self, path: str):
        dirn = os.path.dirname(path)
        # n = os.path.basename(path)
        trash = os.path.join(dirn, '.trash')
        os.makedirs(trash, exist_ok=True)
        shutil.move(path, trash)

    def _move_all_to_trash(self, file_list: List[str]):
        for f in file_list:
            self._move_to_trash(f)

    def delete_migration_files(self, start: int, end: int):
        """Delete migration files by index.

        Args:
            start (int): start index
            end (int): end index
        """
        print(f'=> Deleting migration files for model {self.model.__name__}')
        for i in range(start, end+1):
            si = str(i)
            si = '0' * (self.index_length - len(si)) + si
            files = glob.glob(os.path.join(self.migration_dir, f'{self.db_table}_{si}_*.json'))
            self._move_all_to_trash(files)
            pypat = os.path.join(self.migration_queue_dir, f'{self.db_table}_{si}_*.py')
            pyrpat = b'[^\x00]*' + bytes(f'{self.db_table}_{si}_', encoding='utf-8') + b'[^\x00]*\\.py\x00'
            pyfiles = glob.glob(pypat)
            self._move_all_to_trash(pyfiles)
            content = b''
            with open(self.migration_cursor_path, 'rb') as f:
                content = f.read()
                content = re.sub(pyrpat, b'', content)
            tmpf = self.migration_cursor_path + '.tmp'
            with open(tmpf, 'wb') as f:
                f.write(content)
            os.rename(tmpf, self.migration_cursor_path)
        print(f'   Migration files for model {self.model.__name__} have been successfully deleted.')

    def _update_migration_cursor(self, path: str):
        with open(self.migration_cursor_path, 'ab') as f:
            cont = bytes(path, encoding='utf-8') + b'\x00'
            f.write(cont)

    def _get_unapplied_migrations(self) -> List[str]:
        mgrpy_files = sorted(glob.glob(self.mgrpy_file_pattern))
        mgr = iter(mgrpy_files)
        prev_files = []
        try:
            with open(self.migration_cursor_path, 'rb') as f:
                content = f.read().split(sep=b'\x00')
                for cursor in content:
                    if not cursor: continue # split includes a last empty element
                    try:
                        file = next(mgr)
                        if cursor != bytes(file, encoding='utf-8'):
                            raise ValueError(f"Migration file path mismatch, expected {str(cursor)} == {file}")
                        prev_files.append(file)
                    except StopIteration:
                        raise ValueError(f"Migration file/s missing after {str(cursor)}")
        except FileNotFoundError:
            pass
        return mgrpy_files[len(prev_files):]

    async def migrate(self, pool: Pool):
        """Run the migrations created by makemigrations beforehand.

        Args:
            pool (Pool): pool object.
        """
        print(f'=> Applying migrations for model {self.model.__name__}')
        files = self._get_unapplied_migrations()
        if not files:
            print('   Nothing to migrate. Did you run makemigrations ?')
            return None
        for file in files:
            mn = os.path.basename(file).replace('.py','')
            mr = import_from_path(mn, file) # type: ignore
            async with Transaction(pool) as tdb:
                mro: MigrationRunner = mr.MigrationRunner(tdb, self.model)
                await mro.run()
                self._update_migration_cursor(file)
                print(f'   Migration applied: {mn}')

    def make_migrations(self, yes=False, silent=False):
        """Prepare migration files.

        Args:
            yes (bool, optional): confirm yes to all. Defaults to False.
            silent (bool, optional): Suppress message. Defaults to False.
        """
        print(f'=> Making migrations for model {self.model.__name__} ...')
        query = ''
        qs = []
        if self.pfields:
            # changes only
            for q, m in self._migration_query_generator():
                if not q: continue
                if not silent or not yes:
                    print(m)
                if not yes:
                    self._take_yn()
                qs.append(q)
            query = '\n'.join(qs)
        else:
            # new table
            query = self.get_create_table_query()
            print(query)
            if not yes:
                self._take_yn()
        if not query:
            if not silent:
                print(f"   No changes detected for model {self.model.__name__}")
            return None
        with open(self.current_mgrpy_file_path, 'w') as f:
            f.write(MIGRATION_RUNNER_TEMPLATE.format(migration_query=query))
            with open(self.current_file_path, 'w') as jf:
                # print(self.current_json)
                json.dump(self.current_json, jf)
        print(f'*** Migrations for model {self.model.__name__} have been successfully created.')

    def _migration_query_generator(self) -> Iterator[Tuple[str, str]]:
        """Detect changes on model fields and yield query, discriptive message

        Yields:
            Iterator[Tuple[str, str]]: yield query, descriptive_message
        """
        changed = _get_changed_fields(self.cfields, self.pfields)
        for k, v in changed.items():
            op = v['op']
            if op == 'add':
                query, msg = v['cur_def'].get_query_column_add(self.db_table) # type: ignore
            elif op == 'delete':
                query, msg = v['pre_def'].get_query_column_drop(self.db_table) # type: ignore
            elif op == 'mod':
                query, msg = v['cur_def'].get_query_column_modify(v['pre_def'], self.db_table) # type: ignore
            elif op == 'rename':
                rnm_query, rnm_msg = v['cur_def'].get_query_column_rename(v['pre_key'], self.db_table) # type: ignore
                mod_query, mod_msg = v['cur_def'].get_query_column_modify(v['pre_def'], self.db_table) # type: ignore
                query = f'{rnm_query} {mod_query}'
                msg = f"{mod_msg}\n{rnm_msg}"
            else:
                raise ValueError(f"Invalid operation: {op}")
            yield query, f'\n{"*"*79}{msg}\n\n{query}\n{"*"*79}'


    def _get_previous_migration_file_path(self) -> str:
        """Get latest migration file.

        Raises:
            IndexError: when no files found

        Returns:
            str: file path
        """
        files = sorted(glob.glob(self.migration_file_pattern))
        if files:
            return files[-1]
        return ''

    def _get_current_migration_file_index(self, previous_file_name: str) -> int:
        if not previous_file_name:
            return 1
        m = re.match(f'^{self.db_table}_0*(\\d+)_.*\\.json$', previous_file_name)
        try:
            return int(m.group(1)) + 1  # type: ignore
        except AttributeError:
            raise ValueError(f"Invalid migration file name: {previous_file_name}")

    def _get_current_migration_file_name_without_extension(self, previous_file_name: str) -> str:
        cindex = str(self._get_current_migration_file_index(previous_file_name))
        cindex = '0' * (self.index_length - len(cindex)) + cindex
        timestamp = str(datetime.datetime.now())
        timestamp = re.sub(r'[.:\s-]', '_', timestamp)
        return f"{self.db_table}_{cindex}_{timestamp}"


async def _run_with_transaction(pool: Pool, func: Callable, *args: Any, **kwargs: Any):
    await func(pool, *args, **kwargs)

def run_with_transaction(pool: Pool, func: Callable, *args: Any, **kwargs: Any):
    asyncio.get_event_loop().run_until_complete(_run_with_transaction(pool, func, *args, **kwargs))


def migration_manager(pool: Pool, base_path: str, models: List[ModelType]):
    """Migration manager.

    Parses arguments to run migration commands.

    Args:
        pool (Pool): pool object.
        base_path (str): migration base path
        models (List[ModelType]): models to migrate
    """
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("cmd", type=str,
                        help="""Command:
                        makemigrations: Create migration files.
                        migrate: Apply migrations created by makemigrations command.
                        delete_migration_files <sart> <end>: Delete migration files from start index to end index.""")
    parser.add_argument('start_index', nargs='?', type=int, default=0, help='Start index for delete_migration_files command')
    parser.add_argument('end_index', nargs='?', type=int, default=0, help='End index for delete_migration_files command')
    parser.add_argument('-y', '--yes', action='store_true', help='Confirm all', default=False)
    parser.add_argument('-q', '--quiet', action='store_true', help='Suppress message', default=False)

    args = parser.parse_args()

    if args.cmd == 'makemigrations':
        for model in models:
            Migration(model, base_path).make_migrations(yes=args.yes, silent=args.quiet)
    elif args.cmd == 'migrate':
        for model in models:
            run_with_transaction(pool, Migration(model, base_path).migrate)
    elif args.cmd == 'delete_migration_files':
        if args.start_index == 0:
            raise ValueError(f'start_index and end_index must be given')
        if args.start_index > args.end_index:
            raise ValueError(f'E: Invalid start ({args.start_index}) and end index ({args.end_index})\n')
        for model in models:
            Migration(model, base_path).delete_migration_files(args.start_index, args.end_index)

    else:
        raise ValueError('E: Invalid command: {args.cmd}')



"""Contains the Model's base Meta class
"""

__author__ = 'Md Jahidul Hamid <jahidulhamid@yahoo.com>'
__copyright__ = 'Copyright © Md Jahidul Hamid <https://github.com/neurobin/>'
__license__ = '[BSD](http://www.opensource.org/licenses/bsd-license.php)'
__version__ = '0.0.1'



class MetaType(type):
    def __setattr__(self, k, v):
        raise NotImplementedError("Meta class attribute can not be set outside of class definition.")
    def __delattr__(self, k):
        raise NotImplementedError("Meta class attribute can not be deleted outside of class definition.")


class Meta(metaclass=MetaType):
    """Skeleton class for metadata

    Available meta settings:

    * `db_table` (*str*): db table name,
    * `abstract` (*bool*): Whether it is an abstract model. Abstract models do not have db table and are used as base models.
    * `pk` (*str*):  Primary key. Defaults to 'id',
    * `proxy` (*bool*): Whether it is a proxy model. Defaults to False. Proxy models inherit everything. This is only to have different pythonic behavior of a model. Proxy models can not define new fields and they do not have separate db table but share the same db table as their parents. Proxy setting is always inherited by child model, thus If you want to turn a child model non-proxy, set the proxy setting in its Meta class.
    * `ordering` (*Tuple[str]*): Ordering. Example: `('name', '-price')`, where name is ascending and price is in descending order.
    * `fields_up` (*Tuple[str]*): These fields only will be taken to update or save data onto db. Empty tuple means no restriction.
    * `fields_down` (*Tuple[str]*): These fields only will be taken to select/retrieve data from db. Empty tuple means no restriction.
    * `exclude_fields_up` (*Tuple[str]*): Exclude these fields when updating data to db. Empty tuple means no restriction.
    * `exclude_fields_down` (*Tuple[str]*): Exclude these fields when retrieving data from db. Empty tuple means no restriction.
    * `exclude_values_up` (*Dict[str, Tuple[Any]]*): Exclude fields with these values when updating. Empty dict and empty tuple means no restriction. Example: `{'': (None,), 'price': (0,)}` when field name is left empty ('') that criteria will be applied to all fields.
    * `exclude_values_down` (*Dict[str, Tuple[Any]]*): Exclude fields with these values when retrieving data. Empty dict and empty tuple means no restriction. Example: `{'': (None,), 'price': (0,)}` when field name is left empty ('') that criteria will be applied to all fields.
    * `f`: Access field names.
    """
    # _field_defs_ = {} This must not be included in Meta class
    # If it is included here, Meta class will inherently be allowed to pass
    # this/these fields which is not expected.

    # f is a reserved attribute to access field names
    def __init__(self):
        raise NotImplementedError("Creating instances of Meta class is not supported")




import asyncio
import unittest
from copy import copy, deepcopy
import pickle
from typing import Dict, List, Tuple, Any
from morm.types import Void, VoidType
import morm.migration as mg
from morm.model import Model, ModelType, Field
from morm.fields.field import ColumnConfig
import os, shutil, sys
from morm.db import DB, Pool, Transaction
import morm.exceptions as ex
import random

DB_POOL = Pool(
            dsn='postgres://',
            host='localhost',
            port=5432,
            user='jahid',
            password='jahid',
            database='test',
            min_size=10,
            max_size=90,
)

class TestMethods(unittest.TestCase):

    @classmethod
    async def _asetup(cls):
        cls.DB_POOL = DB_POOL
        cls.mgpath = '/tmp/_morm_mgr_x_' + str(random.random())
        db = DB(cls.DB_POOL)
        await db.execute(f'DROP TABLE IF EXISTS "User";')

    @classmethod
    def setUpClass(cls):
        asyncio.get_event_loop().run_until_complete(cls._asetup())

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.mgpath)

    def setUp(self):
        pass

    def tearDown(self):
        # self.DB_POOL.close() # atexit calls it
        pass


    def test_func(self):

        curs = {
            'name': 'varchar(255)',
            'profession': 'varchar(255)',
            'age': 'int',
            'hobby': 'varchar(255)',
            'height': 'int',
            'weight': 'int',
        }
        pres = {
            'profession': 'varchar(25)',
            'age': 'int',
            'address': 'varchar(255)',
            'hobby': 'varchar(255)',
            'height_': 'int2',
            'weight': 'int',


        }
        # all_altdefs = mg._get_changed_fields(curs, pres)
        # print('\n')
        # for k, v in all_altdefs.items():
        #     sql, msg = _get_alter_column_sql(self, v)
        #     print(msg)
        #     print(sql)
        class User(Model):
            id = Field('SERIAL', sql_onadd='NOT NULL')
            name = Field('varchar(255)')
            profession = Field('varchar(65)', sql_alter=("ALTER TABLE \"{table}\" ALTER COLUMN \"{column}\" SET DEFAULT 'Teacher'",'ALTER TABLE "{table}" ALTER COLUMN "{column}" SET NOT NULL'))

        mgo = mg.Migration(User, '/tmp/..Non-Existent')
        print(' - [x] Field init checks: 1. repr implementation 2. default values')
        # print(repr(User.Meta._field_defs_['name']))
        self.assertTrue( "Field('varchar(255)', sql_onadd='', sql_ondrop='', sql_alter=('ALTER TABLE \"{table}\" DROP CONSTRAINT IF EXISTS \"__UNQ_{table}_{column}__\";',), sql_engine='postgresql'," in repr(User.Meta._field_defs_['name']))

        print(' - [x] Checking pfields and cfields')
        # print(mgo.cfields)
        cfields = {'id': ColumnConfig(sql_type='SERIAL', sql_onadd='NOT NULL', sql_ondrop='', sql_alter=('ALTER TABLE "{table}" DROP CONSTRAINT IF EXISTS "__UNQ_{table}_{column}__";',), sql_engine='postgresql', column_name='id'), 'name': ColumnConfig(sql_type='varchar(255)', sql_onadd='', sql_ondrop='', sql_alter=('ALTER TABLE "{table}" DROP CONSTRAINT IF EXISTS "__UNQ_{table}_{column}__";',), sql_engine='postgresql', column_name='name'), 'profession': ColumnConfig(sql_type='varchar(65)', sql_onadd='', sql_ondrop='', sql_alter=('ALTER TABLE "{table}" DROP CONSTRAINT IF EXISTS "__UNQ_{table}_{column}__";', 'ALTER TABLE "{table}" ALTER COLUMN "{column}" SET DEFAULT \'Teacher\'', 'ALTER TABLE "{table}" ALTER COLUMN "{column}" SET NOT NULL'), sql_engine='postgresql', column_name='profession')}
        self.assertTrue(mgo.cfields == cfields)
        self.assertTrue({} == mgo.pfields)

        mgq = mgo._migration_query_generator()

        print(' - [x] Checking add column sql for id')
        query, msg = next(mgq)
        self.assertTrue('''
*******************************************************************************
* > ADD: id: SERIAL
* + ALTER TABLE "User" DROP CONSTRAINT IF EXISTS "__UNQ_User_id__";

ALTER TABLE "User" ADD COLUMN "id" SERIAL NOT NULL;
ALTER TABLE "User" DROP CONSTRAINT IF EXISTS "__UNQ_User_id__";;
*******************************************************************************''' == msg)
        self.assertTrue(query == 'ALTER TABLE "User" ADD COLUMN "id" SERIAL NOT NULL;\nALTER TABLE "User" DROP CONSTRAINT IF EXISTS "__UNQ_User_id__";;')

        print(' - [x] Checking add column sql for name')
        query, msg = next(mgq)
        # print(msg)
        self.assertTrue('''
*******************************************************************************
* > ADD: name: varchar(255)
* + ALTER TABLE "User" DROP CONSTRAINT IF EXISTS "__UNQ_User_name__";

ALTER TABLE "User" ADD COLUMN "name" varchar(255) ;
ALTER TABLE "User" DROP CONSTRAINT IF EXISTS "__UNQ_User_name__";;
*******************************************************************************''' == msg)
        self.assertTrue(query == 'ALTER TABLE "User" ADD COLUMN "name" varchar(255) ;\nALTER TABLE "User" DROP CONSTRAINT IF EXISTS "__UNQ_User_name__";;')

        print(' - [x] Checking add column sql for profession')
        query, msg = next(mgq)
        # print(msg)
        self.assertEqual('''
*******************************************************************************
* > ADD: profession: varchar(65)
* + ALTER TABLE "User" DROP CONSTRAINT IF EXISTS "__UNQ_User_profession__";
* + ALTER TABLE "User" ALTER COLUMN "profession" SET DEFAULT 'Teacher'
* + ALTER TABLE "User" ALTER COLUMN "profession" SET NOT NULL

ALTER TABLE "User" ADD COLUMN "profession" varchar(65) ;
ALTER TABLE "User" DROP CONSTRAINT IF EXISTS "__UNQ_User_profession__";; ALTER TABLE "User" ALTER COLUMN "profession" SET DEFAULT 'Teacher'; ALTER TABLE "User" ALTER COLUMN "profession" SET NOT NULL;
*******************************************************************************''', msg)
        self.assertEqual(query, """ALTER TABLE "User" ADD COLUMN "profession" varchar(65) ;
ALTER TABLE "User" DROP CONSTRAINT IF EXISTS "__UNQ_User_profession__";; ALTER TABLE "User" ALTER COLUMN "profession" SET DEFAULT 'Teacher'; ALTER TABLE "User" ALTER COLUMN "profession" SET NOT NULL;""")

        print(mgo.get_create_table_query())
        self.assertEqual(mgo.get_create_table_query(), """
CREATE TABLE "User" (
    "id" SERIAL NOT NULL,
    "name" varchar(255) ,
    "profession" varchar(65) @
);ALTER TABLE "User" DROP CONSTRAINT IF EXISTS "__UNQ_User_id__";;
ALTER TABLE "User" DROP CONSTRAINT IF EXISTS "__UNQ_User_name__";;
ALTER TABLE "User" DROP CONSTRAINT IF EXISTS "__UNQ_User_profession__";; ALTER TABLE "User" ALTER COLUMN "profession" SET DEFAULT 'Teacher'; ALTER TABLE "User" ALTER COLUMN "profession" SET NOT NULL;""".replace('@', '').strip())

        mgo = mg.Migration(User, self.mgpath)
        mgo.make_migrations(yes=True)

        class User(Model):
            id = Field('SERIAL', sql_onadd='NOT NULL')
            name = Field('varchar(256)')
            profession_name = Field('varchar(265)', sql_alter=("ALTER TABLE \"{table}\" ALTER COLUMN \"{column}\" SET DEFAULT 'Teacher'",'ALTER TABLE "{table}" ALTER COLUMN "{column}" SET NOT NULL'))
            hobby = Field('varchar(45)')

        mgo = mg.Migration(User, self.mgpath)
        mg.Migration(User, self.mgpath).make_migrations(yes=True)
        mg.Migration(User, self.mgpath).make_migrations(yes=True)
        mg.Migration(User, self.mgpath).make_migrations(yes=True)

        self._migrate(mg.Migration(User, self.mgpath))
        self._migrate(mg.Migration(User, self.mgpath))

        mg.Migration(User, self.mgpath).delete_migration_files(1, 1)

        class AbstractUser(User):
            class Meta:
                abstract = True
        class ProxyUser(User):
            class Meta:
                proxy = True
        with self.assertRaises(ex.MigrationModelNotAllowedError):
            mgo = mg.Migration(AbstractUser, self.mgpath)

        with self.assertRaises(ex.MigrationModelNotAllowedError):
            mgo = mg.Migration(ProxyUser, self.mgpath)

        self.assertEqual(mgo.default_json, mgo._get_json_from_file('___non_existence'))

        class User(Model):
            id = Field('SERIAL', sql_onadd='NOT NULL')
            name = Field('varchar(25)')
            profession_name = Field('varchar(265)', sql_alter=("SET DEFAULT 'Teacher'",'SET NOT NULL'))
        sys.argv = [__file__, 'makemigrations', '-y']
        mg.migration_manager(self.DB_POOL, self.mgpath, [User])
        sys.argv = [__file__, 'migrate']
        mg.migration_manager(self.DB_POOL, self.mgpath, [User])
        sys.argv = [__file__, 'delete_migration_files', '1', '1']
        mg.migration_manager(self.DB_POOL, self.mgpath, [User])
        sys.argv = [__file__, 'delete_migration_file', '1', '1']
        with self.assertRaises(ValueError):
            mg.migration_manager(self.DB_POOL, self.mgpath, [User])
        sys.argv = [__file__, 'delete_migration_files']
        with self.assertRaises(ValueError):
            mg.migration_manager(self.DB_POOL, self.mgpath, [User])
        sys.argv = [__file__, 'delete_migration_files', '2', '1']
        with self.assertRaises(ValueError):
            mg.migration_manager(self.DB_POOL, self.mgpath, [User])


    def _migrate(self, mgo: mg.Migration):
        async def _my_migrate():
            await mgo.migrate(self.DB_POOL)
        asyncio.get_event_loop().run_until_complete(_my_migrate())




if __name__ == "__main__":
    unittest.main(verbosity=0)



import asyncio
import logging
import unittest
import random
from uuid import uuid4
import inspect

from morm.db import Pool, DB, Transaction
import morm.model as mdl
import morm.meta as mt
from morm.types import Void
import morm.fields.field as fdl


LOGGER_NAME = 'morm-test-model-'
log = logging.getLogger(LOGGER_NAME)

def get_file_content(path):
    cont = ''
    try:
        with open(path, 'r') as f:
            cont = f.read();
    except Exception as e:
        log.exception("E: could not read file: " + path)
    return cont


SNORM_DB_POOL = Pool(
    dsn='postgres://',
    host='localhost',
    port=5432,
    user='jahid',
    password='jahid',
    database='test',
    min_size=10,
    max_size=90,
)

from morm.model import Model
from morm.fields import Field

def mprint(*args, **kwargs):
    print("-"*80)
    print(*args, **kwargs)
    print("-"*80)

class TestMethods(unittest.TestCase):

    def checkModelAttributeSet(self, name, value):
        print(f"> set attribute '{name}' for model class must produce NotImplementedError")
        with self.assertRaises(NotImplementedError):
            class User(Model):pass
            setattr(User, name, value)


    def checkModelAttributeDel(self, name):
        print(f"> del attribute '{name}' for model class must produce NotImplementedError")
        with self.assertRaises(NotImplementedError):
            class User(Model):pass
            delattr(User, name)

    def checkMetaAttributeSet(self, name, value):
        print(f"> set attribute '{name}' for Meta class of model must produce NotImplementedError")
        with self.assertRaises(NotImplementedError):
            class User(Model):
                class Meta(mdl.Meta): pass
                setattr(Meta, name, value)

    def checkMetaAttributeDel(self, name, value):
        print(f"> del attribute '{name}' for Meta class of model must produce NotImplementedError")
        with self.assertRaises(NotImplementedError):
            class User(Model):
                class Meta(mdl.Meta): pass
                delattr(Meta, name)

    def test_Model_Attr_Set_Del(self):
        attrs = ['Meta', 'name', 'dummy', 'hozoborolo']
        for k in attrs:
            self.checkModelAttributeSet(k, 32)
            self.checkModelAttributeDel(k)

    def test_Model_Meta(self):
        print("> Meta can not be anything other than a class")
        with self.assertRaises(TypeError):
            class User(Model):
                Meta = 32

    def test_Model_Meta_Internal_fields(self):
        print("> _field_defs_ internal meta field can not be set by user.")
        with self.assertRaises(ValueError):
            class User(Model):
                class Meta(mdl.Meta):
                    _field_defs_ = 67576

    def test_Model_Meta_Attr_Set_Del(self):

        meta_attrs = {
            'pk': 23,
            'proxy': 23,
            'ordering': 23,
            'fields_up': 23,
            'fields_down': 23,
            'exclude_fields_up': 23,
            'exclude_fields_down': 23,
            'exclude_values_up': 23,
            'exclude_values_down': 23,
            'db_table': 23,
            'abstract': 23,
        }

        for k, v in meta_attrs.items():
            self.checkMetaAttributeSet(k, v)
            self.checkMetaAttributeDel(k, v)

    def checkMetaAttributeType(self, name, value):
        print(f"> set attribute '{name}' with invalid (typed) value {value} for Meta class in Model must produce TypeError")
        with self.assertRaises(TypeError):
            class User(Model):
                Meta = mt.MetaType('Meta', (mt.Meta,), {name: value})

    def test_Model_Meta_Attr_Types(self):

        meta_attrs = {
            'pk': 23,
            'proxy': 23,
            'ordering': 23,
            'fields_up': 23,
            'fields_down': 23,
            'exclude_fields_up': 23,
            'exclude_fields_down': 23,
            'exclude_values_up': 23,
            'exclude_values_down': 23,
            'db_table': 23,
            'abstract': 23,
        }

        for k, v in meta_attrs.items():
            self.checkMetaAttributeType(k, v)

    def test_Model_Meta_Attr_Defaults(self):
        class User(Model):
            profession = Field('varchar(255)')
            name = Field('varchar(255)')
        class BigUser(User):
            age = Field("int")

        print("> db_table default value must be the class name")
        self.assertEqual(User.Meta.db_table, 'User')
        self.assertEqual(BigUser.Meta.db_table, 'BigUser')

        print("> Meta._field_defs_ must be a dict")
        self.assertTrue(isinstance(User.Meta._field_defs_, dict))

        meta_attrs_defaults = {
            'pk': 'id',
            'abstract': False,
            'proxy': False,
            'ordering': (),
            'fields_up': (),
            'fields_down': (),
            'exclude_fields_up': (),
            'exclude_fields_down': (),
            'exclude_values_up': {'':()},
            'exclude_values_down': {'':()},
        }
        for k, v in meta_attrs_defaults.items():
            gv = getattr(BigUser.Meta, k)
            print(f"> User Model without Meta class must have default value {v} for Meta attribute {k}")
            self.assertEqual(gv, v)




    def test_Model_Meta_Attribute_Definition(self):
        class User(Model):
            class Meta:
                pk = 'column_id'
                db_table = 'myapp_model_user'
                abstract = True
                proxy = False
                ordering = ('id', 'name')
                fields_up = ('id', 'name')
                fields_down = ('id', 'name')
                exclude_fields_up = ('updated_at',)
                exclude_fields_down = ('password',)
                exclude_values_up = {'': ('', None)}
                exclude_values_down = {'':('', None)}

        meta_attr_inh = {
            'pk': 'column_id',
            'proxy': False,
            'ordering': ('id', 'name'),
            'fields_up': ('id', 'name'),
            'fields_down': ('id', 'name'),
            'exclude_fields_up': ('updated_at',),
            'exclude_fields_down': ('password',),
            'exclude_values_up': {'': ('', None)},
            'exclude_values_down': {'': ('', None)},
        }

        meta_attrs_no_inh = {
            'db_table': Void,
            'abstract': True,
        }

        meta_attrs = {**meta_attr_inh, **meta_attrs_no_inh}

        for k, v in meta_attrs.items():
            print(f"> Meta attribute value for '{k}' must match with defined custom value: {v}")
            self.assertEqual(getattr(User.Meta, k), v)

        # checking Meta inheritance
        class User2(User):
            class Meta:
                pk = 'column_id'
                proxy = False
                ordering = ('id', 'name')
                fields_up = ('id', 'name')
                fields_down = ('id', 'name')
                exclude_fields_up = ('updated_at',)
                exclude_fields_down = ('password',)
                exclude_values_up = {'': ('', None)}
                exclude_values_down = {'': ('', None)}

        class User3(User): pass


        for k, v in meta_attr_inh.items():
            print(f"> Meta attribute value for '{k}' must match with inherited value: {v}")
            self.assertEqual(getattr(User3.Meta, k), v)


        meta_attrs_no_inh_default = {
            'db_table': 'User3',
            'abstract': False,
        }
        for k, v in meta_attrs_no_inh.items():
            mv = getattr(User3.Meta, k)
            dv = meta_attrs_no_inh_default[k]
            print(f"> Meta attribute value for '{k}' must not be inherited and match the default value: {dv}")
            self.assertEqual(mv, dv)

        ############################
        # check proxy
        ############################

        meta_attr_inh = {
            'pk': 'column_id',
            'proxy': True,
            'ordering': ('id', 'name'),
            'fields_up': ('id', 'name'),
            'fields_down': ('id', 'name'),
            'exclude_fields_up': ('updated_at',),
            'exclude_fields_down': ('password',),
            'exclude_values_up': {'': ('', None)},
            'exclude_values_down': {'': ('', None)},
        }

        meta_attrs_no_inh = {
            'db_table': Void,
            'abstract': True,
        }

        meta_attrs = {**meta_attr_inh, **meta_attrs_no_inh}

        # checking Meta inheritance
        class User2(User):
            class Meta:
                proxy = True

        # class User3(User): pass
        print(' - [x] db_table must be the same for proxy models')
        self.assertEqual(User2.Meta.db_table, User.Meta.db_table)

        for k, v in meta_attr_inh.items():
            print(f"> Proxy:True: Meta attribute value for '{k}' must match with inherited value: {v}")
            self.assertEqual(getattr(User2.Meta, k), v)

        meta_attrs_no_inh_default = {
            'db_table': Void,
            'abstract': True,
        }
        for k, v in meta_attrs_no_inh.items():
            mv = getattr(User2.Meta, k)
            print(f"> Proxy:True: Meta attribute value for '{k}' must be inherited and match the value: {v}")
            self.assertEqual(mv, v)

    def test_Model_Instance(self):
        class User(Model):
            name = Field('varchar(255)')
            profession = Field('varchar(255)')

        class BigUser(Model):
            name = Field('varchar(255)')
            profession = Field('varchar(255)')
            age = Field("int")

        class BigUser2(Model):
            name = Field('varchar(255)')
            profession = Field('varchar(255)')
            age = Field("int")

        user = BigUser()
        user.name = 'ffdsf'
        user.age = 34
        user.profession = 'Teacher'
        print("> set attribute on model instance for non-existence field must produce AttributeError")
        with self.assertRaises(AttributeError):
            user.profesion = 'Teacher' # spelling mistake

        print("> del attribute on model instance is allowed")
        del user.profession
        # mprint(inspect.getsource(BigUser2))
        # mprint(user)

    def test_Model_Proxy(self):
        class User(Model):
            name = Field('varchar(255)')
            profession = Field('varchar(255)')
            age = Field("int")
        print("> Proxy model can not include new fields")
        with self.assertRaises(ValueError):
            class Me(User):
                fdslkjlds = Field('varchar(255)')
                name = 34
                class Meta(mdl.Meta):
                    proxy = True

    def test_Model_Field_Order(self):
        class User(Model):
            name = Field('varchar(255)')
            profession = Field('varchar(255)')
            age = Field("int")

        keys = ['name', 'profession', 'age']
        ckey = list(User.Meta._field_defs_.keys())
        print("> Class attribute definition order must be preserved (requires python 3.7+)")
        self.assertEqual(keys, ckey)

    def test_Model_Field(self):
        def name_default():
            return random.choice(User.NAMES)

        class User(Model):
            NAMES = ['John Doe', 'Jane Doe']
            name = Field('varchar(255)', default=name_default)
            profession = Field('varchar(255)', default='Teacher')
            age = Field("int")

        user = User()
        self.assertTrue(user.name in User.NAMES)
        user.name = 'Jahidul Hamid'
        self.assertEqual('Jahidul Hamid', user.name)
        print("> Field().name represents the string name of the field")
        self.assertTrue(User.Meta._field_defs_['name'].name == 'name')

        print("> Model instance is iterable (k,v)")
        for k, v in user:
            print((k, v,))

        with self.assertRaises(AttributeError):
            class User2(Model):
                NAMES = ['John Doe', 'Jane Doe']
                __name__ = Field('varchar(255)', default=name_default)

    def test_Model_Init(self):

        class User(Model):
            NAMES = ['John Doe', 'Jane Doe']
            name = Field('varchar(255)', default='John Doe')
            profession = Field('varchar(255)', default='Teacher')
            age = Field("int", default=23)
            hobby = Field('varchar(255)', default='Gardenning')
            status = Field('varchar(255)', default='NOK')
            marital_status = Field('varchar(255)', default='Single')
            address = Field('varchar(255)', default='Gardenning Park')
            class Meta:
                fields_up = ('name', 'profession', 'age', 'status', 'marital_status', 'address')
                exclude_fields_up = ('profession',)
                exclude_values_up = {
                    '': ('Single',),
                    'address': ('Gardenning Park',),
                }
                fields_down = ('name', 'profession', 'age', 'hobby', 'status', 'marital_status', 'address')
                exclude_fields_down = ('age',)
                exclude_values_down = {
                    '': ('NOK',),
                    'address': (None,),
                }

        user = User({'name': 'Student'}, name='Jahid')
        with self.assertRaises(TypeError):
            user = User(['name', 'Student'], name='Jahid')
        with self.assertRaises(AttributeError):
            dct = user.fdsfdss
        del user.__dict__ # for coverage to touch __delattr__

        # value = fdl.FieldValue(Field('varchar(255)', default='something'))
        # nokvalue = fdl.FieldValue(Field('varchar(255)', default='something'))
        # nokvalue.value = 'NOK'
        # singlevalue = fdl.FieldValue(Field('varchar(255)', default='something'))
        # singlevalue.value = 'Single'
        # print('\n#'*10)
        # print(value.value)
        user = User()
        data = user.Meta._fields_

        ans = [
            ('name', data['name']),
            ('profession', data['profession']),
            ('hobby', data['hobby']),
            ('marital_status', data['marital_status']),
            ('address', data['address']),
        ]
        c = 0
        for k, v in User._get_FieldValue_data_valid_(data, up=False):
            self.assertIn((k, v), ans)
            c += 1
        self.assertEqual(c, len(ans))

        ans = [
            ('name', data['name']),
            ('age', data['age']),
            ('status', data['status']),
        ]
        c = 0
        for k, v in User._get_FieldValue_data_valid_(data, up=True):
            self.assertIn((k, v), ans)
            c += 1
        self.assertEqual(c, len(ans))

        ans = ['name', 'profession', 'hobby', 'status', 'marital_status', 'address']
        c = 0
        for k in User._get_fields_():
            self.assertIn(k, ans)
            c += 1
        self.assertEqual(c, len(ans))

        ans = ['name', 'age', 'status', 'marital_status', 'address']
        c = 0
        for k in User._get_fields_(up=True):
            self.assertIn(k, ans)
            c += 1
        self.assertEqual(c, len(ans))



        # b = BigUser(name='__dummy__', age=23)

    # async def _test_transaction_setup(self):
    #     b = BigUser(name='__dummy__', age=23)
    #     # await b._save_()

    # async def _test_transaction(self):
    #     # try:
    #     #     async with Transaction(BigUser._db_) as con:
    #     #         b = await BigUser._get_(where="name='__dummy__'", con=con)
    #     #         b.age += 2
    #     #         await b._save_(con=con)
    #     #         # raise Exception
    #     # except:
    #     #     pass
    #     # b = await BigUser._get_(where="name='__dummy__'")
    #     # self.assertEqual(b.age, 23)
    #     pass

    # def clean(self):
    #     SNORM_DB_POOL.close()

    # def test_transaction(self):
    #     try:
    #         asyncio.get_event_loop().run_until_complete(self._test_transaction_setup())
    #         # group = asyncio.gather(self._test_transaction_setup(), *[self._test_transaction() for i in range(10)])
    #         # group = asyncio.gather( *[self._test_transaction() for i in range(10000)])
    #         # group = asyncio.gather(self._test_transaction_setup())
    #         # asyncio.get_event_loop().run_until_complete(asyncio.gather(group))
    #     finally:
    #         self.clean()

    # def _test_something(self):
        # db = DB(SNORM_DB_POOL)
        # dbpool = await db.pool()
        # await dbpool.execute('CREATE TABLE IF NOT EXISTS "BigUser" (id SERIAL not null PRIMARY KEY, name varchar(255), profession varchar(255), age int)')
        # # await db.execute('INSERT into test_table')
        # mos = await test_table._get_(where='name like $1 order by id asc', args=['%dumm%'])
        # print(mos.__dict__)
        # # print(mos[0].__dict__)
        # print(BigUser()._fields_)
        # print(BigUser._get_table_name_())
        # b = BigUser()
        # b.name = 'jahid'
        # b.age = 28
        # print(b._get_insert_query_())
        # id = await b._insert_()
        # print(id)

        # b1 = await BigUser._get_(where='id=1')
        # b1.name = 'Jahidul Hamid'
        # print(b1._get_update_query_())
        # await b1._update_()
        # b = BigUser()
        # b.name = 'John Doeee'
        # b.profession = 'Teacher'
        # await b._save_()
        # b.age = 34
        # await b._save_()
        # d = {
        #     'name': 'John Doe',
        #     'age': 45,
        # }
        # b = BigUser(name='Jahid')


if __name__ == "__main__":
    try:
        unittest.main(verbosity=2)
    except:
        # SNORM_DB_POOL.close()
        raise



import logging
import unittest
import os, sys
import tempfile
import shutil
from morm.admin import main


LOGGER_NAME = 'morm-test-admin-'
log = logging.getLogger(LOGGER_NAME)



class TestMethods(unittest.TestCase):
    def test_main(self):
        tmpdir = tempfile.mkdtemp()
        os.chdir(tmpdir)
        sys.argv = [__file__, 'init']
        files = main()
        main() # touch FileExistsError
        for file in files:
            path = os.path.join(tmpdir, file)
            print(f' - [x] Check if {path} content matches default {file} content')
            with open(path, 'r', encoding='utf-8') as f:
                self.assertEqual(files[file], f.read())


        print(f' - [x] Invalid command produces ValueError')
        sys.argv = [__file__, 'int']
        with self.assertRaises(ValueError):
            main()

        shutil.rmtree(tmpdir)




if __name__ == "__main__":
    unittest.main(verbosity=2)



import asyncio
import logging
import unittest
import random
from uuid import uuid4
import inspect

from morm.db import Pool, DB, Transaction
import morm.model as mdl
import morm.meta as mt
import morm.fields.field as fdl
from morm.types import Void
import morm.exceptions as ex


LOGGER_NAME = 'morm-test-field-'
log = logging.getLogger(LOGGER_NAME)



class TestMethods(unittest.TestCase):
    def test_default_funcs(self):
        print("> always_valid returns True all the time")
        self.assertTrue(fdl.always_valid(False) is True)
        print("> nomodifiy returns the exact same value")
        v = 34
        self.assertTrue(v is fdl.nomodify(v))

    def test_Field_Defaults(self):
        f = fdl.Field('varchar(23)')
        print("> Fields default validator is always_valid")
        self.assertTrue(f.validator is fdl.always_valid)
        print("> Fields default modifier is nomodify")
        self.assertTrue(f.modifier is fdl.nomodify)
        print("> Fields default value is Void")
        self.assertTrue(f.default is Void)
        print("> Fields fallback default is False")
        self.assertTrue(f.fallback is False)
        print("> Fields initial value is Void (same as default value)")
        with self.assertRaises(AttributeError):
            self.assertTrue(f.value is Void) # new change

    def test_FieldValue(self):
        v = 45
        f = fdl.FieldValue(fdl.Field("varchar(34)", default='34', validator=lambda x: type(x) is int, modifier=lambda x: int(x)))
        f1 = fdl.FieldValue(fdl.Field("varchar(34)", default='34', validator=lambda x: type(x) is int))
        f2 = fdl.FieldValue(fdl.Field("varchar(34)", default='34', fallback=True, validator=lambda x: type(x) is int))
        print("> default value does not pass through validator")
        self.assertTrue(isinstance(f.value, str))
        with self.assertRaises(ValueError):
            f.value = '454f'

        f.value = '345'

        with self.assertRaises(ValueError):
            f1.value = '34'

        f2.value = '3434' # invalid
        #but fallback is True, so it will not produce exception
        #instead it will be set to default value.
        self.assertTrue(f2.value == '34')
        self.assertTrue(f2.value_change_count == 1)
        f2.value = 3432
        self.assertTrue(f2.value == 3432)
        self.assertTrue(f2.value_change_count == 2)

    def test_ColumnConfig(self):
        cc = fdl.ColumnConfig(sql_engine='mysql', sql_type='int', sql_onadd='', sql_alter=('ALTER TABLE "{table}" ALTER COLUMN "{column}" SET DEFAULT 1',), sql_ondrop='', column_name='name')
        cc2 = fdl.ColumnConfig(sql_engine='mysql', sql_type='varchar', sql_onadd='NOT NULL', sql_alter=('ALTER TABLE "{table}" ALTER COLUMN "{column}" SET DEFAULT 0',), sql_ondrop='CASCADE', column_name='old_name')
        self.assertEqual(repr(cc), '''ColumnConfig(sql_engine='mysql', sql_type='int', sql_onadd='', sql_alter=('ALTER TABLE "{table}" ALTER COLUMN "{column}" SET DEFAULT 1',), sql_ondrop='', column_name='name')''')
        with self.assertRaises(ex.UnsupportedError):
            cc.get_query_column_add('dummy_table')
        with self.assertRaises(ex.UnsupportedError):
            cc.get_query_column_drop('dummy_table')
        with self.assertRaises(ex.UnsupportedError):
            cc.get_query_column_rename('old_name','dummy_table')
        with self.assertRaises(ex.UnsupportedError):
            cc.get_query_column_modify(cc2, 'dummy_table')
        with self.assertRaises(ex.UnsupportedError):
            cc.get_query_column_alter((),'dummy_table')

        cc.conf['sql_engine'] = 'postgresql'
        cc2.conf['sql_engine'] = 'postgresql'
        print(cc.get_query_column_modify(cc2, 'dummy_table'))
        self.assertEqual(
            cc.get_query_column_modify(cc2, 'dummy_table'),
            ('ALTER TABLE "dummy_table" ALTER COLUMN "name" SET DATA TYPE int;\nALTER TABLE "dummy_table" ALTER COLUMN "name" SET DEFAULT 1;', '\n* > MODIFY: old_name: varchar --> int\n* + ALTER TABLE "dummy_table" ALTER COLUMN "name" SET DEFAULT 1')
            )

        cc.conf['sql_alter'] = 'SET SOMETHING'
        with self.assertRaises(TypeError):
            cc.get_query_column_alter(cc2.conf['sql_alter'], 'dummy_table')
        cc.conf['sql_alter'] = ('SET SOMETHING',)
        cc2.conf['sql_alter'] = 'SET SOMETHING'
        with self.assertRaises(TypeError):
            cc.get_query_column_alter(cc2.conf['sql_alter'], 'dummy_table')

        name = fdl.Field('varchar(234)')
        name2 = fdl.Field('varchar(232)')
        self.assertTrue(name != name2)



if __name__ == "__main__":
    unittest.main(verbosity=2)



"""Model.
"""

__author__ = 'Md Jahidul Hamid <jahidulhamid@yahoo.com>'
__copyright__ = 'Copyright © Md Jahidul Hamid <https://github.com/neurobin/>'
__license__ = '[BSD](http://www.opensource.org/licenses/bsd-license.php)'
__version__ = '0.0.1'

import inspect
import typing
from typing import Optional, Dict, List, Tuple, TypeVar, Union, Any, Iterator
from collections import OrderedDict
import copy
from abc import ABCMeta
from asyncpg import Record # type: ignore
from morm.exceptions import ItemDoesNotExistError
from morm.fields.field import Field, FieldValue
from morm.types import Void
import morm.meta as mt      # for internal use

# morm.db must not be imported here.

Meta = mt.Meta  # For client use


class _FieldNames():
    """Access field names
    """
    def __init__(self, func):
        self.__dict__['func'] = func

    def __getattr__(self, k):
        return self.__dict__['func'](k)

    def __setattr__(self, k, v):
        raise NotImplementedError


class ModelType(type):
    Meta: typing.ClassVar # fixing mypy error: "ModelType" has no attribute "Meta"
    def __new__(mcs, class_name: str, bases: tuple, attrs: dict):
        # Ensure initialization is only performed for subclasses of Model
        # excluding Model class itself.
        parents = tuple(b for b in bases if isinstance(b, ModelType))
        if not parents:
            return super().__new__(mcs, class_name, bases, attrs)

        classcell = attrs.pop('__classcell__', None)
        class _Meta_(mt.Meta): pass
        meta = attrs.pop('Meta', _Meta_)
        if not inspect.isclass(meta): #TEST: Meta is restricted as a class
            raise TypeError(f"Name 'Meta' is reserved for a class to pass configuration or metadata of a model. Error in model '{class_name}'")
        _class_ = super().__new__(mcs, 'x_' + class_name, parents, attrs)
        BaseMeta = getattr(_class_, 'Meta', _Meta_)

        meta_attrs = {}
        def _set_meta_attr(k, v, mutable=False, inherit=True, internal=False):
            try:
                given_value = getattr(meta, k)
                if internal:
                    raise ValueError(f"'{k}' is a reserved attribute for class Meta. Error in model '{class_name}'")
                given_type = type(given_value)
                required_type = type(v)
                if not given_type is required_type:
                    raise TypeError(f"Invalid type {given_type} given for attribute '{k}' in class '{class_name}.Meta'. Required {required_type}.")
                meta_attrs[k] = given_value
            except AttributeError:
                if inherit:
                    v = getattr(BaseMeta, k, v)
                # mutable values can be changed by other class meta change
                if mutable:
                    meta_attrs[k] = copy.deepcopy(v)
                else:
                    meta_attrs[k] = v

        _set_meta_attr('proxy', False)
        _set_meta_attr('pk', 'id')
        _set_meta_attr('ordering', ())
        _set_meta_attr('fields_up', ())
        _set_meta_attr('fields_down', ())
        _set_meta_attr('exclude_fields_up', ())
        _set_meta_attr('exclude_fields_down', ())
        _set_meta_attr('exclude_values_up', {'':()}, mutable=True)
        _set_meta_attr('exclude_values_down', {'':()}, mutable=True)
        _set_meta_attr('_field_defs_', {}, internal=True, mutable=True)

        if meta_attrs['proxy']:
            #proxy model inherits everything
            try:
                meta_attrs['db_table'] = BaseMeta.db_table
                meta_attrs['abstract'] = BaseMeta.abstract
            except AttributeError:
                raise TypeError(f"This model '{class_name}' can not be a proxy model. It does not have a valid base or super base non-proxy model")
        else:
            _set_meta_attr('abstract', False, inherit=False)
            if meta_attrs['abstract']:
                meta_attrs['db_table'] = Void
            else:
                _set_meta_attr('db_table', class_name, inherit=False)

        new_attrs = {}

        # dict is ordered, officially from python 3.7
        for n, v in _class_.__dict__.items():
            if isinstance(v, Field):
                if n.startswith('_'):
                    raise AttributeError(f"Invalid field name '{n}' in model '{class_name}'. \
                        Field name must not start with underscore.")
                if meta_attrs['proxy'] and n in attrs:
                    raise ValueError(f"Proxy model '{class_name}' can not define new field: {n}")
                v.name = n
                # v.sql_conf.conf['table_name'] = meta_attrs['db_table'] # Field must not contain table_name, because it is void when model is abstract and it gets inherited.
                meta_attrs['_field_defs_'][n] = v
            elif n in attrs:
                new_attrs[n] = attrs[n]

        # we do this after finalizing meta_attr
        def _get_field_name(n: str) -> str:
            if n in meta_attrs['_field_defs_']:
                return n
            else:
                raise AttributeError(f"No such field '{n}' in model '{class_name}'")
        meta_attrs['f'] = _FieldNames(_get_field_name)

        MetaClass = mt.MetaType('Meta', (mt.Meta,), meta_attrs)
        new_attrs['Meta'] = MetaClass

        if classcell is not None:
            new_attrs['__classcell__'] = classcell
        return super().__new__(mcs, class_name, bases, new_attrs)

    def __repr__(cls):
        return inspect.getsource(cls)

    def __setattr__(self, k, v):
        raise NotImplementedError("You can not set model attributes outside model definition.")

    def __delattr__(self, k):
        raise NotImplementedError("You can not delete model attributes outside model definition.")

    def _is_valid_key_(self, k:str, fields:Tuple[str], exclude_keys:Tuple[str]) -> bool:
        """Returns True if the key is valid considering include/exclude keys
        """
        if k in exclude_keys: return False
        if fields and k not in fields: return False
        return True

    def _is_valid_down_key_(self, k: str) -> bool:
        """Returns True if the key is valid considering include/exclude down keys
        """
        return self._is_valid_key_(k, self.Meta.fields_down, self.Meta.exclude_fields_down)

    def _is_valid_up_key_(self, k: str) -> bool:
        """Returns True if the key is valid considering include/exclude up keys
        """
        return self._is_valid_key_(k, self.Meta.fields_up, self.Meta.exclude_fields_up)

    def _is_valid_value_(self, k: str, v: Any, exclude_values: Dict[str, Tuple[Any]]) -> bool:
        """Returns True if the value for the key is valid considering exclude values
        """
        if v is Void:
            return False
        if k in exclude_values:
            if v in exclude_values[k]:
                return False
        if '' in exclude_values and v in exclude_values['']:
            return False
        return True

    def _is_valid_up_value_(self, k: str, v: Any) -> bool:
        """Returns True if the value for the key is valid considering exclude up values
        """
        return self._is_valid_value_(k, v, self.Meta.exclude_values_up)

    def _is_valid_down_value_(self, k: str, v: Any) -> bool:
        """Returns True if the value for the key is valid considering exclude down values
        """
        return self._is_valid_value_(k, v, self.Meta.exclude_values_down)

    def _is_valid_down_(self, k: str, v: Any) -> bool:
        """Check whether the key and value is valid for down (data retrieval)
        """
        return self._is_valid_down_key_(k) and self._is_valid_down_value_(k, v)

    def _is_valid_up_(self, k: str, v: Any) -> bool:
        """Check whether the key and value is valid for up (data update)
        """
        return self._is_valid_up_key_(k) and self._is_valid_up_value_(k, v)

    def _get_all_fields_(self) -> Dict[str, Field]:
        """Get all fields on model without applying any restriction.

        Returns:
            Dict[str, Field]: Dictionary of all fields
        """
        return self.Meta._field_defs_

    def _get_all_fields_json_(self) -> Dict[str, Dict]:
        """Get all fields on model without applying any restriction in JSON like dict

        Returns:
            Dict[str, Dict]: Dictionary of all fields in JSON like dict
        """
        res = {}
        all_fields = self._get_all_fields_()
        for k,v in all_fields.items():
            res[k] = v.to_json()
        return res

    def _check_field_name_(self, n: str) -> str:
        """Return the field name if exists else raise AttributeError

        Args:
            n (str): field name

        Raises:
            AttributeError: if field name does not exist

        Returns:
            str: field name
        """
        if n in self.Meta._field_defs_:
            return n
        else:
            raise AttributeError(f"No such field `{n}` in model `{self.__name__}`")

    def _get_fields_(self, up=False) -> Iterator[str]:
        """Yields field names that pass include/exclude criteria

        Args:
            up (bool, optional): up criteria or down criteria. Defaults to False (down).

        Yields:
            str: field name
        """
        if up:
            fields = self.Meta.fields_up
            exclude_keys = self.Meta.exclude_fields_up
        else:
            fields = self.Meta.fields_down
            exclude_keys = self.Meta.exclude_fields_down
        all_fields = self._get_all_fields_()
        for k in all_fields:
            if not self._is_valid_key_(k, fields, exclude_keys):
                continue
            yield k

    def _get_fields_json_(self, up=False) -> Dict[str, Dict]:
        """Get fields in JSON like dict that pass include/exclude criteria

        Args:
            up (bool, optional): up criteria or down criteria. Defaults to False (down).

        Returns:
            Dict[str, Dict]: Dictionary of fields in JSON like dict
        """
        res = {}
        fields = self._get_fields_(up)
        all_fields = self._get_all_fields_()
        for k in fields:
            res[k] = all_fields[k].to_json()
        return res

    def _get_FieldValue_data_valid_(self, data: dict, up=False) -> Iterator[Tuple[str, Any]]:
        """Yields valid key,value pairs from data.

        Validity is checked against include/exclude key/value criteria.

        Args:
            data (dict): data to be validated.
            up (bool, optional): whether up (data update) or down (data retrieval). Defaults to False.

        Yields:
            Iterator[Tuple[str, Any]]: Yields key, value pair
        """
        if up:
            exclude_values = self.Meta.exclude_values_up
            fields = self.Meta.fields_up
            exclude_fields = self.Meta.exclude_fields_up
        else:
            exclude_values = self.Meta.exclude_values_down
            fields = self.Meta.fields_down
            exclude_fields = self.Meta.exclude_fields_down
        # new_data = type(data)()
        for k,v in data.items():
            if not self._is_valid_key_(k, fields, exclude_fields):
                continue
            if not self._is_valid_value_(k, v.value, exclude_values):
                continue
            yield k, v


    # def _get_data_for_valid_values_(self, data, up=False, gen=False):
    #     if up:
    #         exclude_values = self.Meta.exclude_values_up
    #     else:
    #         exclude_values = self.Meta.exclude_values_down
    #     new_data = type(data)()
    #     for k,v in data.items():
    #         if not self._is_valid_value_(k, v, exclude_values):
    #             continue
    #         if gen:
    #             yield k, v
    #         else:
    #             new_data[k] = v
    #     if not gen:
    #         return new_data

    def _get_db_table_(self) -> str:
        """Get db table name for model
        """
        return self.Meta.db_table

    def _is_abstract_(self) -> bool:
        """Whether it's an abstract model or not
        """
        return self.Meta.abstract

    def _is_proxy_(self) -> bool:
        """Whether it is a proxy model or not
        """
        return self.Meta.proxy

    def _get_pk_(self) -> str:
        """Get primary column name
        """
        return self.Meta.pk

    def _get_ordering_(self, quote: str) -> Iterator[Tuple[str, str]]:
        """Yield each ordering from model parsed and converted to column, direction

        direction is either `ASC` or `DESC`

        Args:
            quote (str): Quote to apply to the column

        Yields:
            Iterator[Tuple[str, str]]: Yields column, direction
        """
        ordering = self.Meta.ordering
        direction = 'ASC'
        for o in ordering:
            if o.startswith('-'):
                direction = 'DESC'
                o = o[1:]
            elif o.startswith('+'):
                o = o[1:]
            o = f"{quote}{o}{quote}"
            yield o, direction


class ModelBase(metaclass=ModelType):
    """Base Model for all models.

    Do not inherit from this class, use Model instead.

    Raises:
        TypeError: When invalid type is encountered
        AttributeError: When misspelled fields are tried to set.
    """
    class Meta:
        """Meta that holds metadata for model
        """
        # The following needs to be defined here, not in meta.Meta
        # meta.Meta is used in client Models, thus everything
        # included there will be blindly inherited, while these are passed
        # through the metaclasses __new__ methods and processed accordingly
        # to determine which one should be inherited and which one should not.
        pk = 'id'
        '''Primary key'''
        db_table = Void
        abstract = True
        proxy = False
        ordering = ()
        fields_up = ()
        fields_down = ()
        exclude_fields_up = ()
        exclude_fields_down = ()
        exclude_values_up = {'':()}
        exclude_values_down = {'':()}

        #internal
        _field_defs_: Dict[str, Field]
        _fields_: Dict[str, FieldValue]
        _fromdb_: List[str]


    def __init__(self, *args, **kwargs):
        class Meta:
            _fields_: Dict[str, FieldValue] = {}
            _fromdb_: List[str] = []
        # super(ModelBase, self).__setattr__('Meta', Meta)
        self.__dict__['Meta'] = Meta
        for k, v in self.__class__.Meta._field_defs_.items():
            self.Meta._fields_[k] = FieldValue(v)
        for arg in args:
            try:
                arg_items = arg.items()
            except AttributeError:
                raise TypeError(f"Invalid argument type ({type(arg)}) to Model __init__ method. Expected: dictionary or keyword argument")
            for k,v in arg_items:
                setattr(self, k, v)
        for k,v in kwargs.items():
            setattr(self, k, v)

    def __iter__(self):
        """Iter through k, v where k is field name and v is field value

        Yields:
            tuple: field_name, field_value
        """
        for k, f in self.Meta._fields_.items():
            if self.__class__._is_valid_down_(k, f.value):
                yield k, f.value

    def __delattr__(self, k):
        fields = self.Meta._fields_
        if k in fields:
            fields[k].delete_value()
        else:
            super().__delattr__(k)

    def __getattr__(self, k):
        Meta = self.__dict__['Meta']
        fields = Meta._fields_
        if k in fields:
            v = fields[k].value
            if self.__class__._is_valid_down_(k, v):
                return v
            raise AttributeError(f'Invalid attempt to access field `{k}`. It is excluded using either exclude_fields_down or exclude_values_down in {self.__class__.__name__} Meta class. Or it does not have any valid value.')
        raise AttributeError

    def __setattr__(self, k: str, v):
        if k == 'Meta':
            raise AttributeError(f"Name '{k} is reserved. You should not try to change it.")
        if k.startswith('_'):
            if k.endswith('_'):
                raise AttributeError('_<name>_ such names are reserved for predefined methods.')
            self.__dict__[k] = v
            return
        fields = self.Meta._fields_
        if k not in fields:
            raise AttributeError(f"No such field ('{k}') in model '{self.__class__.__name__}''")
        # v = fields[k].clean(v)
        # super().__setattr__(k, v)
        if k in self.Meta._fromdb_:
            self.Meta._fromdb_.remove(k)
            if self.__class__._is_valid_down_(k, v):
                fields[k].value = v
            elif self.__class__._is_valid_up_(k, v):
                fields[k]._ignore_first_change_count_ = True
                fields[k].value = v
            else:
                raise AttributeError(f'Can not set field `{k}`. It is excluded using either exclude_fields_up/down or exclude_values_up/down in {self.__class__.__name__} Meta class. Or you are trying to set an invalid value: {v}')
        elif self.__class__._is_valid_up_(k, v):
            fields[k].value = v
        else:
            raise AttributeError(f'Can not set field `{k}`. It is excluded using either exclude_fields_up/down or exclude_values_up/down in {self.__class__.__name__} Meta class. Or you are trying to set an invalid value: {v}')

    def __repr__(self):
        reprs = []
        for k, v in self:
            reprs.append(f'{k}={repr(v)}')
        body = ', '.join(reprs)
        return f'{self.__class__.__name__}({body})'


    async def _pre_save_(self, db):
        """Pre-save hook.

        Override to run pre save cleanup.

        Args:
            db (DB): db handle.
        """
        pass

    async def _pre_delete_(self, db):
        """Pre-delete hook.

        Override to run pre delete cleanup.

        Args:
            db (DB): db handle.
        """
        pass


    async def _post_save_(self, db):
        """Pre-save hook.

        Override to run post save cleanup.

        Args:
            db (DB): db handle.
        """
        pass

    async def _post_delete_(self, db):
        """Pre-delete hook.

        Override to run post delete cleanup.

        Args:
            db (DB): db handle.
        """
        pass



    async def _pre_insert_(self, db):
        """Pre-insert hook.

        Override to run pre insert cleanup.

        Args:
            db (DB): db handle.
        """
        pass

    async def _pre_update_(self, db):
        """Pre-update hook.

        Override to run pre update cleanup.

        Args:
            db (DB): db handle.
        """
        pass


    async def _post_insert_(self, db):
        """Pre-insert hook.

        Override to run post insert cleanup.

        Args:
            db (DB): db handle.
        """
        pass

    async def _post_update_(self, db):
        """Pre-update hook.

        Override to run post update cleanup.

        Args:
            db (DB): db handle.
        """
        pass



class Model(ModelBase):
    """Base model to be inherited by other models.

    It's more than a good practice to define a Base model first:

    ```python
    from morm.model import Model
    from morm.datetime import timestamp

    class Base(Model):
        class Meta:
            pk = 'id' # setting primary key, it is defaulted to 'id'
            abstract = True

        # postgresql example
        id = Field('SERIAL', sql_onadd='PRIMARY KEY NOT NULL')
        created_at = Field('TIMESTAMP WITH TIME ZONE', sql_onadd='NOT NULL', sql_alter=('ALTER TABLE "{table}" ALTER COLUMN "{column}" SET DEFAULT NOW()',))
        updated_at = Field('TIMESTAMP WITH TIME ZONE', sql_onadd='NOT NULL', value=timestamp)
    ```

    Then a minimal model could look like this:

    ```python
    class User(Base):
        name = Field('varchar(65)')
        email = Field('varchar(255)')
        password = Field('varchar(255)')
    ```

    An advanced model could look like:

    ```python
    import random

    def get_rand():
        return random.randint(1, 9)

    class User(Base):
        class Meta:
            db_table = 'myapp_user'
            abstract = False    # default is False
            proxy = False       # default is False
            # ... etc...
            # see morm.meta.Meta for supported meta attributes.

        name = Field('varchar(65)')
        email = Field('varchar(255)')
        password = Field('varchar(255)')
        profession = Field('varchar(255)', default='Unknown')
        random = Field('integer', default=get_rand) # function can be default
    ```

    ## Initialize a model instance

    keyword arguments initialize corresponding fields according to
    the keys.

    Positional arguments must be dictionaries of
    keys and values.

    Example:

    ```python
    User(name='John Doe', profession='Teacher')
    User({'name': 'John Doe', 'profession': 'Teacher'})
    User({'name': 'John Doe', 'profession': 'Teacher'}, age=34)
    User({'name': 'John Doe', 'profession': 'Teacher', 'active': True}, age=34)
    ```

    Raises:
        TypeError: If invalid type of argument is provided.

    ## Special Model Meta attribute `f`:

    You can access field names from `ModelClass.Meta.f`.

    This allows a spell-safe way to write the field names. If you
    misspell the name, you will get `AttributeError`.

    ```python
    f = User.Meta.f
    my_data = {
        f.name: 'John Doe',         # safe from spelling mistake
        f.profession: 'Teacher',    # safe from spelling mistake
        'hobby': 'Gardenning',      # unsafe from spelling mistake
    }
    ```
    """
    class Meta:
        # The following needs to be defined here, not in meta.Meta
        # meta.Meta is used in client Models, thus everything
        # included there will be blindly inherited, while these are passed
        # through the metaclasses __new__ methods and processed accordingly
        # to determine which one should be inherited and which one should not.
        abstract = True

    def __init__(self, *args, **kwargs):
        super(Model, self).__init__(*args, **kwargs)



"""Common Postgresql Models.
"""

__author__ = 'Md Jahidul Hamid <jahidulhamid@yahoo.com>'
__copyright__ = 'Copyright © Md Jahidul Hamid <https://github.com/neurobin/>'
__license__ = '[BSD](http://www.opensource.org/licenses/bsd-license.php)'
__version__ = '0.0.1'


from morm.model import Model
from morm.fields import Field
from morm.datetime import timestamp


class Base(Model):
    """A model that defines default column 'id'
    """
    class Meta:
        abstract = True
    id = Field('SERIAL', sql_onadd='PRIMARY KEY NOT NULL', help_text='Auto incremented integer')


class BaseCommon(Base):
    """A model that defines default columns:

    id: Auto incremented integer
    created_at: timestamp
    update_at: timestamp (used with pythonic default)
    """
    class Meta:
        abstract = True
    created_at = Field('TIMESTAMP WITH TIME ZONE', sql_onadd='NOT NULL', sql_alter=('ALTER TABLE "{table}" ALTER COLUMN "{column}" SET DEFAULT NOW()',), help_text='Timestamp of creation, defaults to NOW()')
    updated_at = Field('TIMESTAMP WITH TIME ZONE', sql_onadd='NOT NULL', value=timestamp, help_text='Timestamp of last update, auto updated to NOW()')



"""datetime handling.
"""

__author__ = 'Md Jahidul Hamid <jahidulhamid@yahoo.com>'
__copyright__ = 'Copyright © Md Jahidul Hamid <https://github.com/neurobin/>'
__license__ = '[BSD](http://www.opensource.org/licenses/bsd-license.php)'
__version__ = '0.0.1'


from datetime import datetime, timezone


def timestamp() -> datetime:
    """Get current timestamp.

    Returns:
        datetime.datetime
    """
    return datetime.now()

def timestampu() -> datetime:
    """Get current UTC timestamp.

    Example: '2021-03-12 05:29:22.497195+00:00'

    Returns:
        datetime.datetime
    """
    return datetime.now(timezone.utc)



"""DB utilities.
"""

__author__ = 'Md Jahidul Hamid <jahidulhamid@yahoo.com>'
__copyright__ = 'Copyright © Md Jahidul Hamid <https://github.com/neurobin/>'
__license__ = '[BSD](http://www.opensource.org/licenses/bsd-license.php)'
__version__ = '0.1.0'

import collections
import re
import asyncio
import nest_asyncio  # type: ignore
import atexit
import logging

import asyncpg # type: ignore
from asyncpg import Record, Connection # type: ignore
from typing import Optional, Dict, List, Tuple, TypeVar, Union, Any

from morm import exceptions
from morm.model import ModelType, Model, ModelBase, _FieldNames
from morm.q import Q
from morm.types import Void

LOGGER_NAME = 'morm.db-'
log = logging.getLogger(LOGGER_NAME)


nest_asyncio.apply()

def record_to_model(record: Record, model_class: ModelType) -> Model:
    """Convert a Record object to Model object.

    Args:
        record (Record): Record object.
        model_class (ModelType): Model class

    Returns:
        Model: Model instance.
    """
    new_record = model_class()
    for k,v in record.items():
        new_record.Meta._fromdb_.append(k)
        setattr(new_record, k, v)
    return new_record


class Pool(object):
    """Open database connection pool.

    ```python
    from morm.db import Pool

    DB_POOL = Pool(
        dsn='postgres://',
        host='localhost',
        port=5432,
        user='jahid',       # change accordingly
        password='jahid',   # change accordingly
        database='test',    # change accordingly
        min_size=10,        # change accordingly
        max_size=90,        # change accordingly
    )
    ```

    This will create and open an asyncpg pool which will be automatically closed at exit.

    You should set this in a settings file from where you can import the `DB_POOL`

    Args:
        dsn (str, optional): DSN string. Defaults to None.
        min_size (int, optional): Minimum connection in the pool. Defaults to 10.
        max_size (int, optional): Maximum connection in the pool. Defaults to 100.
        max_queries (int, optional): Number of queries after a connection is closed and replaced with a new connection. Defaults to 50000.
        max_inactive_connection_lifetime (float, optional): Number of seconds after which inactive connections in the pool will be closed.  Pass `0` to disable this mechanism. Defaults to 300.0.
        setup ([type], optional): A coroutine to prepare a connection right before it is returned  from `Pool.acquire()`. Defaults to None.
        init ([type], optional): A coroutine to initialize a connection when it is created. Defaults to None.
        loop ([type], optional): Asyncio even loop instance. Defaults to None.
        connection_class ([type], optional): The class to use for connections.  Must be a subclass of `asyncpg.connection.Connection`. Defaults to asyncpg.connection.Connection.
    """
    def __init__(self, dsn: str = None,
                 min_size: int = 10,
                 max_size: int = 100,
                 max_queries: int = 50000,
                 max_inactive_connection_lifetime: float = 300.0,
                 setup=None,
                 init=None,
                 loop=None,
                 connection_class=Connection,
                 **connect_kwargs):
        self.dsn = dsn
        self.min_size = min_size
        self.max_size = max_size
        self.max_queries = max_queries
        self.max_inactive_connection_lifetime = max_inactive_connection_lifetime
        self.setup = setup
        self.init = init
        self.loop = loop
        self.connection_class = connection_class
        self.connect_kwargs = connect_kwargs

        self._pool = None
        self._open()
        atexit.register(self._close)

    @property
    def pool(self) -> asyncpg.pool.Pool:
        """Property pool that is an asyncpg.pool.Pool object
        """
        return self._pool

    async def __create_pool(self) -> asyncpg.pool.Pool:
        """Create a asyncpg.pool.Pool for this Pool object.

        Returns:
            asyncpg.pool.Pool: Pool object (singleton)
        """
        return await asyncpg.create_pool(
                                        dsn=self.dsn,
                                        min_size=self.min_size,
                                        max_size=self.max_size,
                                        max_queries=self.max_queries,
                                        max_inactive_connection_lifetime=self.max_inactive_connection_lifetime,
                                        setup=self.setup,
                                        init=self.init,
                                        loop=self.loop,
                                        connection_class=self.connection_class,
                                        **self.connect_kwargs)
    def _open(self):
        """Open the pool. Called on init so not need to call this
        method explicitly.
        """
        if not self._pool:
            self._pool = asyncio.get_event_loop().run_until_complete(self.__create_pool())
            log.debug("Pool opened")

    def _close(self):
        """Attempt to close the pool gracefully. registered with atexit.
        You do not need to call this method explicitly.
        """
        if self._pool:
            asyncio.get_event_loop().run_until_complete(self._pool.close())
            self._pool = None
            log.debug("Pool closed")


class DB(object):
    """Initialize a DB object setting a pool to get connection from.

    If connection is given, it is used instead.

    The `corp()` method returns an asyncpg.pool.Pool or an
    asyncpg.Connection

    Args:
        pool (Pool): A connection pool
        con (Connection): Connection. Defaults to None.
    """

    def __init__(self, pool: Pool, con: Connection=None):
        self._pool = pool
        self._con = con
        self.DATA_NO_CHANGE = 'DATA_NO_CHANGE_TRIGGERED'

    def corp(self) -> Union[asyncpg.pool.Pool, Connection]:
        """Return the connection if available, otherwise return a Pool.

        Note: The name reads 'c or p'

        Returns:
            asyncpg.Connection or asyncpg.pool.Pool object
        """
        if self._con:
            return self._con
        return self._pool.pool

    async def fetch(self, query: str, *args,
                    timeout: float = None,
                    model_class: ModelType=None
                    ) -> Union[List[ModelBase], List[Record]]:
        """Make a query and get the results.

        Resultant records can be mapped to model_class objects.

        Args:
            query (str): Query string.
            args (*list or *tuple): Query arguments.
            timeout (float, optional): Timeout value. Defaults to None.
            model_class (Model, optional): Defaults to None.

        Returns:
            List[Model] or List[Record] : List of model instances if model_class is given, otherwise list of Record instances.
        """
        pool = self.corp()
        records = await pool.fetch(query, *args, timeout=timeout)
        if not model_class:
            return records
        else:
            new_records = []
            for record in records:
                new_record = record_to_model(record, model_class)
                new_records.append(new_record)
            return new_records

    async def fetchrow(self, query: str, *args,
                        timeout: float = None,
                        model_class: ModelType=None
                        ) -> Union[ModelBase, Record]:
        """Make a query and get the first row.

        Resultant record can be mapped to model_class objects.

        Args:
            query (str): Query string.
            args (*list or *tuple): Query arguments.
            timeout (float, optional): Timeout value. Defaults to None.
            model_class (Model, optional): Defaults to None.

        Returns:
            Record or model_clas object or None if no rows were selected.
        """
        pool = self.corp()
        record = await pool.fetchrow(query, *args, timeout=timeout)
        if not model_class:
            return record
        else:
            if not record:
                return record
            new_record = record_to_model(record, model_class)
            return new_record

    async def fetchval(self, query: str, *args,
                        column: int = 0,
                        timeout: float = None
                        ) -> Any:
        """Run a query and return a column value in the first row.

        Args:
            query (str): Query to run.
            args (*list or *tuple): Query arguments.
            column (int, optional): Column index. Defaults to 0.
            timeout (float, optional): Timeout. Defaults to None.

        Returns:
            Any: Coulmn (indentified by index) value of first row.
        """
        pool = self.corp()
        return await pool.fetchval(query, *args, column=column, timeout=timeout)

    async def execute(self, query: str, *args,
                        timeout: float = None
                        ) -> str:
        """Execute a query.

        Args:
            query (str): Query to run.
            args (*list or *tuple): Query arguments.
            timeout (float, optional): Timeout. Defaults to None.

        Returns:
            str: Status of the last SQL command
        """
        pool = self.corp()
        return await pool.execute(query, *args, timeout=timeout)

    def get_insert_query(self, mob: ModelBase, reset=False) -> Tuple[str, List[Any]]:
        """Get insert query for the model object (mob) with its current data

        Args:
            mob (ModelBase): Model object
            reset (bool): Reset the value change counter. Defaults to False

        Returns:
            (str, list): query, args
        """
        data = mob.Meta._fields_
        new_data_gen = mob.__class__._get_FieldValue_data_valid_(data, up=True)
        columns = []
        values = []
        markers = []
        c = 0
        for n,v in new_data_gen:
            c += 1
            if reset:
                v.value_change_count = 0
                mob.Meta._fromdb_ = []
            columns.append(n)
            values.append(v.value)
            markers.append(f'${c}')
        column_q = '","'.join(columns)
        if column_q:
            column_q = f'"{column_q}"'
            marker_q = ', '.join(markers)
            query = f'INSERT INTO "{mob.__class__._get_db_table_()}" ({column_q}) VALUES ({marker_q}) RETURNING "{mob.__class__._get_pk_()}"'
        else:
            query = ''
        return query, values

    def get_update_query(self, mob: ModelBase, reset=False) -> Tuple[str, List[Any]]:
        """Get the update query for the changed data in the model object (mob)

        Args:
            mob (ModelBase): Model object
            reset (bool): If True, this method can be called just once to get the changes done on mob. Subsequent call will return empty query.

        Raises:
            AttributeError: If primary key does not exists i.e if not updatable

        Returns:
            str, args: tuple of query, args
        """
        pkval = getattr(mob, mob.__class__._get_pk_()) #save method depends on it's AttributeError
        data = mob.Meta._fields_
        new_data_gen = mob.__class__._get_FieldValue_data_valid_(data, up=True)
        colval = []
        values = []
        c = 0
        for n,v in new_data_gen:
            if n == mob.__class__._get_pk_(): continue
            if v.value_change_count > 0:
                c += 1
                colval.append(f'"{n}"=${c}')
                values.append(v.value)
                if reset:
                    v.value_change_count = 0
        colval_q = ', '.join(colval)
        if colval_q:
            where = f'"{mob.__class__._get_pk_()}"=${c+1}'
            values.append(pkval)
            query = f'UPDATE "{mob.__class__._get_db_table_()}" SET {colval_q} WHERE {where}'
        else:
            query = ''
        return query, values

    def get_delete_query(self, mob: ModelBase) -> Tuple[str, List[Any]]:
        """Get the delete query for the model object.

        Args:
            mob (ModelBase): model object.

        Returns:
            Tuple[str, List[Any]]: quey, args
        """
        pkval = getattr(mob, mob.__class__._get_pk_())
        query = f'DELETE FROM "{mob.__class__._get_db_table_()}" WHERE "{mob.__class__._get_pk_()}"=$1'
        return query, [pkval]

    async def delete(self, mob: ModelBase, timeout: float = None) -> str:
        """Delete the model object data from database.

        Args:
            mob (ModelBase): Model object
            timeout (float): timeout value. Defaults to None.

        Returns:
            (str): status of last sql command.
        """
        query, args = self.get_delete_query(mob)
        await mob._pre_delete_(self)
        res = await self.execute(query, *args, timeout=timeout)
        await mob._post_delete_(self)
        return res

    async def insert(self, mob: ModelBase, timeout: float = None) -> Any:
        """Insert the current data state of mob into db.

        Args:
            mob (ModelBase): Model object
            timeout (float): timeout value. Defaults to None.

        Returns:
            (Any): Value of primary key of the inserted row
        """
        query, args = self.get_insert_query(mob, reset=True)
        await mob._pre_insert_(self)
        pkval = await self.fetchval(query, *args, timeout=timeout)
        if pkval is not None:
            setattr(mob, mob.__class__._get_pk_(), pkval)
        await mob._post_insert_(self)
        return pkval

    async def update(self, mob: ModelBase, timeout: float = None) -> str:
        """Update the current changed data of mob onto db

        Args:
            mob (ModelBase): Model object
            timeout (float): timeout value. Defaults to None.

        Raises:
            AttributeError: If primary key does not exists.

        Returns:
            str: status of last sql command.
            Successful status starts with the word 'UPDATE' followed by
            number of rows updated, which should be 1 in this case.
        """
        query, args = self.get_update_query(mob, reset=True)
        if query:
            await mob._pre_update_(self)
            res = await self.execute(query, *args, timeout=timeout)
            await mob._post_update_(self)
        else:
            res = self.DATA_NO_CHANGE
        return res


    async def save(self, mob: ModelBase, timeout: float = None) -> Union[str, Any]:
        """Insert if not exists and update if exists.

        update is tried first, if fails (if pk does not exist), insert
        is called.

        Args:
            mob (ModelBase): Model object
            timeout (float): timeout value. Defaults to None.

        Returns:
            int or str: The value of the primary key for insert or
                            status for update.
        """
        await mob._pre_save_(self)
        try:
            res = await self.update(mob, timeout=timeout)
        except AttributeError:
            res = await self.insert(mob, timeout=timeout)
        await mob._post_save_(self)
        return res

    def q(self, model: ModelType = None) -> 'ModelQuery':
        """Return a ModelQuery for model

        If `None` is passed, it will give a `ModelQuery` without setting
        `self.model` on the `ModelQuery` object.

        Args:
            model (ModelType, optional): model class. Defaults to None.

        Raises:
            TypeError: If invalid model type is passed

        Returns:
            ModelQuery: ModelQuery object
        """
        return self(model)


    def __call__(self, model: ModelType = None) -> 'ModelQuery':
        """Return a ModelQuery for model

        If `None` is passed, it will give a `ModelQuery` without setting
        `self.model` on the `ModelQuery` object.

        Args:
            model (ModelType, optional): model class. Defaults to None.

        Raises:
            TypeError: If invalid model type is passed

        Returns:
            ModelQuery: ModelQuery object
        """
        if isinstance(model, ModelType) or model is None:
            return ModelQuery(self, model)
        raise TypeError(f"Invalid model: {model}. model must be of type {ModelType.__name__}. Make sure you did not pass a model object by mistake.")



class ModelQuery():
    """Query builder for model class.

    Calling `db(Model)` gives you a model query handler which have several query methods to help you make queries.

    Use `q(query, *args)` method to make queries with positional arguments. If you want named arguments, use the uderscored version of these methods. For example, `q(query, *args)` has an underscored version `q_(query, *args, **kwargs)` that can take named arguments.

    You can add a long query part by part:

    ```python
    from morm.db import DB

    db = DB(DB_POOL) # get a db handle.
    qh = db(User)   # get a query handle.

    query, args = qh.q(f'SELECT * FROM {qh.db_table}')\
                    .q(f'WHERE {qh.f.profession} = ${qh.c}', 'Teacher')\
                    .q_(f'AND {qh.f.age} = :age', age=30)\
                    .getq()
    print(query, args)
    # fetch:
    await qh.fetch()
    ```

    The `q` family of methods (`q, qc, qu etc..`) can be used to
    build a query step by step. These methods can be chained
    together to break down the query building in multiple steps.

    Several properties are available to get information of the model
    such as:

    1. `qh.db_table`: Quoted table name e.g `"my_user_table"`.
    2. `qh.pk`: Quoted primary key name e.g `"id"`.
    3. `qh.ordering`: ordering e.g `"price" ASC, "quantity" DESC`.
    4. `qh.f.<field_name>`: quoted field names e.g`"profession"`.
    5. `qh.c`: Current available position for positional argument (Instead of hardcoded `$1`, `$2`, use `f'${qh.c}'`, `f'${qh.c+1}'`).

    `qh.c` is a counter that gives an integer representing the
    last existing argument position plus 1.

    `reset()` can be called to reset the query to start a new.

    To execute a query, you need to run one of the execution methods
    : `fetch, fetchrow, fetchval, execute`.

    **Notable convenience methods:**

    * `qupdate(data)`: Initialize a update query for data
    * `qfilter()`: Initialize a filter query upto WHERE clasue.
    * `get(pkval)`: Get an item by primary key.

    Args:
        db (DB): DB object
        model_class (ModelType): model
    """
    def __init__(self, db: DB, model_class: ModelType = None):
        self.reset()
        self.db = db
        self.model = model_class # can be None
        def func(k):
            return Q(model_class._check_field_name_(k))
        self._f = _FieldNames(func) # no reset

    def __repr__(self):
        return f'ModelQuery({self.db}, {self.model})'

    def reset(self) -> 'ModelQuery':
        """Reset the model query by returning it to its initial state.

        Returns:
            self (Enables method chaining)
        """
        self._query_str_queue: List[str] = []
        self.end_query_str = ''
        self.start_query_str = ''
        self._args: List[Any] = []
        self._arg_count = 0
        self._named_args: Dict[str, Any] = {}
        self._named_args_mapper: Dict[str, int] = {}
        self.__filter_initiated = False
        self._ordering = ''
        self.__update_initiated = False
        return self


    @property
    def c(self) -> int:
        """Current available argument position in the query

        arg_count + 1 i.e if $1 and $2 has been used so far, then
        self.c is 3 so that you can use it to make $3.

        Returns:
            int
        """
        return self._arg_count + 1

    @property
    def db_table(self) -> str:
        """Table name of the model (quoted)
        """
        return Q(self.model._get_db_table_()) #type: ignore

    @property
    def pk(self) -> str:
        """Primary key name (quoted)
        """
        return Q(self.model._get_pk_()) #type: ignore

    @property
    def ordering(self) -> str:
        """Ordering query in SQL, does not include `ORDER BY`.

        Example: `"price" ASC, "quantity" DESC`
        """
        if not self._ordering:
            self._ordering = ','.join([' '.join(y) for y in self.model._get_ordering_(quote='"')]) # type: ignore
        return self._ordering

    @property
    def f(self) -> _FieldNames:
        """Field name container where names are quoted.

        It can be used to avoid spelling mistakes in writing query.

        Example: query `'select "profesion" from "table_name"'`

        will only produce error after actually running the query against
        a correctly spelled column 'profession'.

        while,

        query `f'select {self.f.profesion} from {self.db_table}'`

        will throw python exception telling you that there is no
        misspelled 'profesion' field.

        Note: you have to change `self` in above to the current
        `ModelQuery` instance
        """
        return self._f

    def _process_positional_args(self, *args):
        if args:
            self._args.extend(args)
            self._arg_count += len(args)


    def _process_keyword_args(self, q: str, **kwargs) -> str:
        # TODO: improvents need to be done
        # 1. needs to handle only unquoted keyword :field_name
        #    and ignore ':field_name' or ":field_name"
        self._named_args.update(kwargs)
        for k,v in self._named_args.items():
            if k in self._named_args_mapper:
                q, mc = re.subn(f':{k}\\b', f'${self._named_args_mapper[k]}', q)
            else:
                q, mc = re.subn(f':{k}\\b', f'${self._arg_count+1}', q)
                if mc > 0:
                    self._args.append(v)
                    self._arg_count += 1
                    self._named_args_mapper[k] = self._arg_count
        return q

    def q(self, q: str, *args: Any) -> 'ModelQuery':
        """Add raw query stub without parsing to check for keyword arguments

        Use `$1`, `$2` etc. for arguments.

        Use `self.c` (instance property, use fstring) to get the current
        available argument position.

        This is an efficient way to add query that do not have any
        keyword arguments to handle, compared to `q_()` which checks for
        keyword arguments everytime it is called.

        Example:

        ```python
        mq = db(SomeModel)
        mq\
        .q('SELECT * FROM "table" WHERE $1', True)\
        .q('AND "price" >= $2', 33)\
        .q(f'OR "price" = ${mq.c}', 0) # mq.c=3 (now)\
        .q_('OR "status" = :status', status='OK')\
        # :status is $4:
        .q('OR "active" = $5', 0)\
        .q_('AND "status" = :status')\
        # status='OK' from previous call
        .q('OR "price" = $2')\
        # $2=33 from previous call
        #using format string and mq.c to get the argument position:
        .q(f'OR "price" > ${mq.c} OR "quantity" > ${mq.c+1}', 12, 3)
        #               mq.c=6 ^
        ```

        Args:
            q (str): raw query string
            *args (Any): positional arguments

        Returns:
            ModelQuery: self, enables method chaining.
        """
        self._process_positional_args(*args)
        self._query_str_queue.append(q)
        return self


    def q_(self, q: str, *args, **kwargs) -> 'ModelQuery':
        """Add a query stub having keyword params.

        Use the format `:field_name` for keyword parameter.
        `:field_name` is converted to positional parameter (`$n`).

        This method checks the query against all keyword arguments
        that has been added so far with other `q*()` methods.

        Args:
            q (str): query string (SQL)

        Returns:
            ModelQuery: returns `self` to enable method chaining
        """
        self._process_positional_args(*args)
        q = self._process_keyword_args(q, **kwargs)
        self._query_str_queue.append(q)
        return self

    def qq(self, word: str) -> 'ModelQuery':
        """Quote and add a word to the query.

        Enable to add names with auto-quote. For example, if the name
        for a field value is `status`, it can be added to the query
        with auto-quoting, i.e for postgresql it will be added
        as `"status"`.

        Example:

        ```python
        .qq('price').q('>= $1',34)
        ```

        Args:
            word (str): the word that needs to be added with quote.

        Returns:
            ModelQuery: returns `self` to enable method chaining
        """
        if word:
            self._query_str_queue.append(Q(word))
        return self


    def qc(self, word: str, rest: str, *args) -> 'ModelQuery':
        """Add query by quoting `word` while adding the `rest` as is.

        This is a shorthand for making where clause conditions.
        For example: `qc('price', '>=$1', 34)` is a safe way to write
        a where condition like: `"price" >=34`.

        The same can be achieved by using a combination of
        `qq()` and `q()` or manually quoting and using
        with `q()`

        Example:

        ```python
        .qc('price', '>= $1', 34)
        ```

        Args:
            word (str): left part of query that needs to be quoted
            rest (str): right part of query that does not need to be quoted
            *args (any): args

        Returns:
            ModelQuery: returns `self` to enable method chaining
        """
        return self.qq(word).q(rest, *args)


    def qc_(self, word: str, rest: str, *args, **kwargs) -> 'ModelQuery':
        """Add query by quoting `word` while adding the `rest` as is.

        Same as `qc()` except this method parses the `rest` query string
        for keyword params in the format: `:field_name`

        Args:
            word (str): left part of query that needs to be quoted
            rest (str): right part of query that does not need to be quoted
            *args (any): args
            *kwargs: keyword args

        Returns:
            ModelQuery: returns `self` to enable method chaining
        """
        return self.qq(word).q_(rest, *args, **kwargs)

    def qorder(self):
        """Add ORDER BY

        Returns:
            ModelQuery: returns `self` to enable method chaining
        """
        return self.q('ORDER BY')

    def qo(self, order: str) -> 'ModelQuery':
        """Convert `+/-field_name,` to proper order_by criteria and add to query.

        Example: `-field_name,` will become: `"field_name" DESC,`

        * `+` at beginning means ascending order (default)
        * `-` at beginning means descending order
        * `,` at end means you will add more order criteria

        Ommit the comma (`,`) when it is the last ordering criteria.

        Args:
            order (str): order criteria in the format `+/-field_name,`

        Returns:
            ModelQuery: returns `self` to enable method chaining
        """
        direction = 'ASC'
        if order.startswith('-'):
            order = order[1:]
            direction = 'DESC'
        elif order.startswith('+'):
            order = order[1:]
        if order.endswith(','):
            order = order[0:-1]
            direction += ','
        return self.qq(order).q(direction)

    def qu(self, data: dict) -> 'ModelQuery':
        """Convert data to `"column"=$n` query with args as the
            values and add to the main query.

        The counter of positional arguments increases by the number of
        items in `data`. Make use of `self.c` counter to add more
        queries after using this method.

        Args:
            data (dict): data in format: `{'column': value}`

        Returns:
            ModelQuery: returns `self` to enable method chaining
        """
        setq = ', '.join([f'"{c}"=${i}' for i,c in enumerate(data, self.c)])
        return self.q(setq, *data.values())

    def qreturning(self, *column_names) -> 'ModelQuery':
        """Convenience to add a `RETURNING` clause.

        Args:
            column_names: column names.

        Returns:
            ModelQuery: returns `self` to enable method chaining
        """
        q = '","'.join(column_names)
        if q:
            q = f'RETURNING "{q}"'
        return self.q(q)

    def qwhere(self) -> 'ModelQuery':
        """Convenience to add 'WHERE' to the main query.

        Make use of `qc()` method to add conditions.

        Returns:
            ModelQuery: returns `self` to enable method chaining
        """
        return self.q('WHERE')

    def qfilter(self, no_ordering=False) -> 'ModelQuery':
        """Initiate a filter.

        This initiates a `SELECT` query upto `WHERE`. You can then use the
        `q()`, `qc()`, etc. methods to add conditions and finally
        execute the `fetch()` method to get all results or execute the
        `fetchrow()` method to get a single row.

        Example:

        ```python
        .qfilter().q('"price" >= $1 AND "status" = $2', 32.12, 'OK')
        ```

        Args:
            no_ordering (bool): Whether to remove the default ordering SQL. Defaults to False.

        Returns:
            ModelQuery: returns self to enable method chaining
        """
        if not self.__filter_initiated:
            down_fields = ','.join([Q(x) for x in self.model._get_fields_(up=False)]) #type: ignore
            self.reset().q(f'SELECT {down_fields} FROM "{self.model._get_db_table_()}" WHERE') #type: ignore
            self.__filter_initiated = True
            order_by = self.ordering
            if order_by and not no_ordering:
                self.end_query_str = f'ORDER BY {order_by}'
        else:
            raise ValueError(f"Filter is already initiated for this {self.__class__.__name__} query object: {self}")
        return self

    def qupdate(self, data: dict) -> 'ModelQuery':
        """Initiate a UPDATE query for data.

        This initiates an `UPDATE` query upto `WHERE` and leaves you to
        add conditions with other methods such as `qc` or the generic
        method `q()`.

        Finally call the `execute()` method to execute the query or
        call the `fetchval()` method if using `RETURNING` clause.

        Args:
            data (dict): data in key value dictionary

        Returns:
            ModelQuery: returns `self` to enable method chaining
        """
        if not self.__update_initiated:
            self.reset().q(f'UPDATE {self.db_table} SET').qu(data).qwhere()
            self.__update_initiated = True
        else:
            raise ValueError(f"update is already initiated for this {self.__class__.__name__} query: {self}")
        return self


    def getq(self) -> Tuple[str, List[Any]]:
        """Return query string and arg list

        Returns:
            tuple: (str, list) : (query, args)
        """
        query = ' '.join(self._query_str_queue)
        self._query_str_queue = [query]
        query = f'{self.start_query_str} {query} {self.end_query_str}'
        return query, self._args

    async def fetch(self, timeout: float = None) -> Union[List[ModelBase], List[Record]]:
        """Run query method `fetch` that returns the results in model class objects

        Returns the results in model class objects.

        Args:
            timeout (float, optional): Timeout in seconds. Defaults to None.

        Returns:
            List[Model]: List of model instances.
        """
        query, args = self.getq()
        return await self.db.fetch(query, *args, timeout=timeout, model_class=self.model)

    async def fetchrow(self, timeout: float = None) -> Union[ModelBase, Record]:
        """Make a query and get the first row.

        Resultant record is mapped to model_class object.

        Args:
            timeout (float, optional): Timeout value. Defaults to None.

        Returns:
            model_clas object or None if no rows were selected.
        """
        query, args = self.getq()
        return await self.db.fetchrow(query, *args, timeout=timeout, model_class=self.model)

    async def fetchval(self, column: int = 0, timeout: float = None) -> Any:
        """Run the query and return a column value in the first row.

        Args:
            column (int, optional): Column index. Defaults to 0.
            timeout (float, optional): Timeout. Defaults to None.

        Returns:
            Any: Coulmn (indentified by index) value of first row.
        """
        query, args = self.getq()
        return await self.db.fetchval(query, *args, column=column, timeout=timeout)

    async def execute(self, timeout: float = None) -> str:
        """Execute the query.

        Args:
            timeout (float, optional): Timeout. Defaults to None.

        Returns:
            str: Status of the last SQL command
        """
        query, args = self.getq()
        return await self.db.execute(query, *args, timeout=timeout)

    async def get(self, *vals: Any, col: str = '', comp: str = '=$1') -> Union[ModelBase, Record]:
        """Get the first row found by column and value.

        If `col` is not given, it defaults to the primary key (`pk`) of
        the model.

        If comparison is not given, it defaults to `=$1`

        Example:

        ```python
        from morm.db import DB

        db = DB(DB_POOL) # get a db handle.

        # get by pk:
        user5 = await db(User).get(5)

        # price between 5 and 2000
        user = await db(User).get(5, 2000, col='price', comp='BETWEEN $1 AND $2')
        ```

        Args:
            *vals (any): Values to compare. Must be referenced with $1, $2 etc.. in `comp`.
            col (str, optional): Column name. Defaults to the primary key.
            comp (str, optional): Comparison. Defaults to '=$1'.

        Returns:
            model_clas object or None if no rows were selected.
        """
        if not col:
            col = self.model.Meta.pk    #type: ignore
        return await self.reset().qfilter().qc(col, comp, *vals).fetchrow()

SERIALIZABLE = 'serializable'
REPEATABLE_READ = 'repeatable_read'
READ_COMMITTED = 'read_committed'

class Transaction():
    """Start a transaction.

    Example:

    ```python
    from morm.db import Transaction

    async with Transaction(DB_POOL) as tdb:
        # use tdb just like you use db
        user6 = await tdb(User).get(6)
        user6.age = 34
        await tdb.save(user6)
        user5 = await tdb(User).get(5)
        user5.age = 34
        await tdb.save(user6)
    ```

    Args:
        pool (Pool): Pool instance.
        isolation (str, optional): Transaction isolation mode, can be one of:
            'serializable',
            'repeatable_read',
            'read_committed'.
            Defaults to 'read_committed'.
            See https://www.postgresql.org/docs/9.5/transaction-iso.html
        readonly (bool, optional): Specifies whether this transaction is read-only. Defaults to False.
        deferrable (bool, optional): Specifies whether this transaction is deferrable. Defaults to False.
    """
    def __init__(self, pool: Pool, *,
                isolation: str=READ_COMMITTED,
                readonly: bool=False,
                deferrable: bool=False):
        self._pool = pool
        self.db = DB(None) # type: ignore
        self.tr = None
        self.tr_args = {
            'isolation': isolation,
            'readonly': readonly,
            'deferrable': deferrable,
        }

    async def __aenter__(self) -> DB:
        return await self.start()

    async def start(self) -> DB:
        """Start transaction.

        Raises:
            exceptions.TransactionError: When same object is used simultaneously for transaction

        Returns:
            DB: DB object.
        """
        if self.db._con:
            raise exceptions.TransactionError('Another transaction is running (or not ended properly) with this Transaction object')
        self.db._con = await self._pool.pool.acquire() # type: ignore
        self.tr = self.db._con.transaction(**self.tr_args) # type: ignore
        await self.tr.start() # type: ignore
        return self.db

    async def rollback(self):
        """Rollback the transaction.
        """
        if self.tr:
            await self.tr.rollback()

    async def commit(self):
        """Commit the transaction.
        """
        if self.tr:
            await self.tr.commit()

    async def end(self):
        """Close the transaction gracefully.

        Resources are released and some cleanups are done.
        """
        try:
            if self.db._con:
                await self._pool.pool.release(self.db._con)
        finally:
            self.db._con = None
            self.tr = None

    async def __aexit__(self, extype, ex, tb):
        try:
            if extype is not None:
                await self.rollback()
            else:
                await self.commit()
        finally:
            await self.end()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import os
import sys
from codecs import open
from setuptools import setup, find_packages

sys.path[0:0] = ['morm']

from version import __version__

def get_readme(filename):
    content = ""
    try:
        with open(os.path.join(os.path.dirname(__file__), filename), 'r', encoding='utf-8') as readme:
            content = readme.read()
    except Exception as e:
        pass
    return content

setup(name="morm",
      version=__version__,
      author="Md. Jahidul Hamid",
      author_email="jahidulhamid@yahoo.com",
      description="A minimal asynchronous database object relational mapper",
      license="BSD",
      keywords="async, orm, postgresql",
      url="https://github.com/neurobin/python-morm",
      packages=find_packages(include=('morm*',)),
      long_description=get_readme("README.md"),
      long_description_content_type="text/markdown",
      classifiers=[
        # See: https://pypi.python.org/pypi?:action=list_classifiers
        'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
      ],
      python_requires='>=3.7.0',
      install_requires=['asyncpg','nest_asyncio'],
      scripts=['morm/morm_admin', 'morm/init_fap',],
test_suite="morm.tests.test")
